#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

import copy
import json
from typing import Optional

from PySide6.QtWidgets import QApplication

from pygpt_net.item.assistant import AssistantStoreItem
from pygpt_net.utils import trans


class VectorStore:
    def __init__(self, window=None):
        """
        Assistant vector store editor controller

        :param window: Window instance
        """
        self.window = window
        self.dialog = False
        self.config_initialized = False
        self.current = None
        self.width = 800
        self.height = 500
        self.id = "assistant.store"
        self.options = {
            "id": {
                "type": "text",
                "label": "assistant.store.id",
                "read_only": True,
                "value": "",
            },
            "name": {
                "type": "text",
                "label": "assistant.store.name",
                "value": "",
            },
            "expire_days": {
                "type": "int",
                "label": "assistant.store.expire_days",
                "value": 0,
            },
            "status": {
                "type": "textarea",
                "label": "assistant.store.status",
                "read_only": True,
                "value": "",
            },
        }

    def get_options(self) -> dict:
        """
        Get options dict

        :return: options dict
        """
        return self.options

    def get_option(self, key: str) -> Optional[dict]:
        """
        Get option by key

        :param key: option key
        :return: option dict
        """
        if key in self.options:
            return self.options[key]

    def setup(self):
        """Set up vector store editor"""
        idx = None
        self.window.assistant_store.setup(idx)  # widget dialog setup

    def toggle_editor(self):
        """Toggle vector store editor dialog"""
        if self.dialog:
            self.close()
        else:
            self.open()

    def reset(self):
        """Reset vector store editor"""
        self.current = None
        if self.dialog:
            self.init()

    def open(self, force: bool = False):
        """
        Open vector store editor dialog

        :param force: force open dialog
        """
        if not self.config_initialized:
            self.setup()
            self.config_initialized = True
        if not self.dialog or force:
            self.current = self.window.controller.assistant.editor.get_selected_store_id()
            self.init()
            self.window.ui.dialogs.open(
                "assistant.store",
                width=self.width,
                height=self.height,
            )
            self.dialog = True

    def close(self):
        """Close vector store editor dialog"""
        if self.dialog:
            self.window.ui.dialogs.close('assistant.store')
            self.dialog = False

    def init(self):
        """Initialize vector store editor options"""
        self.reload_items()

        # select the first store if not selected
        if self.current is None:
            self.current = self.get_first_visible()

        # assign store to config dialog fields
        options = copy.deepcopy(self.get_options())  # copy options
        if self.current is not None and self.window.core.assistants.store.has(self.current):
            store = self.window.core.assistants.store.items[self.current]
            data_dict = store.to_dict()
            for key in options:
                if key in data_dict:
                    value = data_dict[key]
                    options[key]["value"] = value
                    if key == "status":
                        options[key]["value"] = json.dumps(value, indent=4)  # as JSON to textarea

            self.set_tab_by_id(self.current)

            # load and apply options to config dialog
            self.window.controller.config.load_options(self.id, options)
        else:
            self.current = None  # reset if not exists
            self.window.controller.config.load_options(self.id, options)

    def refresh_status(self):
        """Reload store status"""
        if self.current is not None:  # TODO: reset on profile reload
            if self.window.core.assistants.store.has(self.current):
                self.window.update_status(trans('status.sending'))
                QApplication.processEvents()
                store = self.window.core.assistants.store.items[self.current]
                self.refresh_store(store)
                self.window.update_status(trans('status.assistant.saved'))
                self.update()  # update stores list in assistant dialog

    def refresh_store(
            self,
            store: AssistantStoreItem,
            update: bool = True
    ):
        """
        Refresh store by ID

        :param store : store object
        :param update: update store after refresh
        """
        # update from API
        self.window.core.assistants.store.update_status(store.id)
        self.window.core.assistants.store.update(store)

        if update and store.id == self.current:
            self.update_current()

    def refresh_by_idx(self, idx: int):
        """
        Refresh store by idx

        :param idx: store idx
        """
        store_id = self.get_by_tab_idx(idx)
        if store_id is not None:
            self.refresh_by_store_id(store_id)

    def refresh_by_store_id(self, store_id: str):
        """
        Refresh store by ID

        :param store_id: store id
        """
        if store_id is not None and store_id in self.window.core.assistants.store.items:
            store = self.window.core.assistants.store.items[store_id]
            if store is not None:
                self.window.update_status(trans('status.sending'))
                QApplication.processEvents()
                self.refresh_store(store)
                self.window.update_status(trans('status.assistant.saved'))
                self.update()

    def update_current(self):
        """Update current store"""
        if self.current is not None and self.window.core.assistants.store.has(self.current):
            store = self.window.core.assistants.store.items[self.current]
            # update textarea
            option = copy.deepcopy(self.get_option("status"))
            option["value"] = json.dumps(store.status, indent=4)
            self.window.controller.config.apply(self.id, "status", option)

            # update name
            option = copy.deepcopy(self.get_option("name"))
            option["value"] = store.name
            self.window.controller.config.apply(self.id, "name", option)

            # update expire days
            option = copy.deepcopy(self.get_option("expire_days"))
            option["value"] = store.expire_days
            self.window.controller.config.apply(self.id, "expire_days", option)

    def save(self, persist: bool = True):
        """
        Save vector store editor

        :param persist: persist to file and close dialog
        """
        if self.current is not None:
            current = self.window.core.assistants.store.items[self.current].to_dict()
            options = copy.deepcopy(self.get_options())  # copy options
            data_dict = {}
            for key in options:
                if key == "status":
                    data_dict[key] = current[key]  # use initial value
                    continue  # skip status
                value = self.window.controller.config.get_value(
                    parent_id="assistant.store",
                    key=key,
                    option=options[key],
                )
                data_dict[key] = value
            self.window.core.assistants.store.items[self.current].from_dict(data_dict)

        # save config
        if persist:
            self.window.update_status(trans('status.sending'))
            QApplication.processEvents()
            if self.current is not None:
                store = self.window.core.assistants.store.update(
                    self.window.core.assistants.store.items[self.current]
                )
                if store is None:
                    self.window.update_status(trans('status.error'))
                    self.window.ui.dialogs.alert("Failed to save vector store")
                    return

            self.update()  # update stores list in assistant dialog
            self.window.update_status(trans("info.settings.saved"))
            self.restore_selection()

    def reload_items(self):
        """Reload list items"""
        items = self.window.core.assistants.store.items
        self.window.assistant_store.update_list("assistant.store.list", items)
        self.restore_selection()

    def restore_selection(self):
        """Restore selection"""
        if self.current is not None:
            idx = self.get_tab_by_id(self.current)
            if idx is not None:
                self.set_by_tab(idx)

    def select(self, idx: int):
        """
        Select store by idx

        :param idx: idx on list
        """
        self.save(persist=False)
        self.current = self.get_by_tab_idx(idx)
        self.init()

    def new(self):
        """Create new vector store"""
        self.window.update_status(trans('status.sending'))
        QApplication.processEvents()

        store = self.window.core.assistants.store.create()
        if store is None:
            self.window.update_status(trans('status.error'))
            self.window.ui.dialogs.alert("Failed to create new vector store")
            return

        self.window.update_status(trans('status.assistant.saved'))

        self.window.core.assistants.store.update(store)
        self.update()  # update stores list in assistant dialog

        # switch to created store
        self.current = store.id
        idx = self.get_tab_by_id(self.current)
        self.set_by_tab(idx)
        self.init()
        self.restore_selection()
        self.refresh_by_store_id(store.id)

    def delete_by_idx(
            self,
            idx: int,
            force: bool = False
    ):
        """
        Delete store by idx

        :param idx: store idx
        :param force: force delete
        """
        store_id = self.get_by_tab_idx(idx)
        self.delete(store_id, force=force)

    def delete(
            self,
            store_id: Optional[str] = None,
            force: bool = False
    ):
        """
        Delete store by idx

        :param store_id: store id
        :param force: force delete
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type="assistant.store.delete",
                id=store_id,
                msg=trans("dialog.assistant.store.delete.confirm"),
            )
            return

        if store_id is None:
            self.window.ui.dialogs.alert("Please select vector store first.")
            return

        self.window.update_status(trans('status.sending'))
        QApplication.processEvents()
        if self.current == store_id:
            self.current = None
        try:
            print("Deleting store: {}".format(store_id))
            if self.window.core.assistants.store.delete(store_id):
                self.window.controller.assistant.batch.remove_store_from_assistants(store_id)
                self.window.update_status(trans('status.deleted'))
                self.window.core.assistants.store.save()
                self.window.controller.assistant.files.update()
                self.update()  # update stores list in assistant dialog
                self.init()
                self.restore_selection()
            else:
                self.window.update_status(trans('status.error'))
        except Exception as e:
            self.window.update_status(trans('status.error'))
            self.window.ui.dialogs.alert(e)

    def set_by_tab(self, idx: int):
        """
        Set current list by tab index

        :param idx: tab index
        """
        store_idx = 0
        for id in self.window.core.assistants.store.get_ids():
            if self.window.core.assistants.store.is_hidden(id):
                continue
            if store_idx == idx:
                self.current = id
                break
            store_idx += 1
        current = self.window.ui.models['assistant.store.list'].index(idx, 0)
        self.window.ui.nodes['assistant.store.list'].setCurrentIndex(current)

    def set_tab_by_id(self, store_id: str):
        """
        Set current list to id

        :param store_id: store id
        """
        idx = self.get_tab_idx(store_id)
        current = self.window.ui.models['assistant.store.list'].index(idx, 0)
        self.window.ui.nodes['assistant.store.list'].setCurrentIndex(current)

    def get_tab_idx(self, store_id: str) -> int:
        """
        Get list index (including hidden)

        :param store_id: model id
        :return: list index
        """
        store_idx = None
        i = 0
        for id in self.window.core.assistants.store.get_ids():
            if self.window.core.assistants.store.is_hidden(id):
                continue
            if id == store_id:
                store_idx = i
                break
            i += 1
        return store_idx

    def get_tab_by_id(self, store_id: str) -> int:
        """
        Get list index (including hidden)

        :param store_id: store id
        :return: list index
        """
        idx = None
        i = 0
        for id in self.window.core.assistants.store.get_ids():
            if self.window.core.assistants.store.is_hidden(id):
                continue
            if id == store_id:
                idx = i
                break
            i += 1
        return idx

    def get_by_tab_idx(self, idx: int) -> Optional[str]:
        """
        Get key by list index (including hidden)

        :param idx: list index
        :return: store id / key
        """
        store_idx = 0
        for id in self.window.core.assistants.store.get_ids():
            if self.window.core.assistants.store.is_hidden(id):
                continue
            if store_idx == idx:
                return id
            store_idx += 1
        return None

    def get_first_visible(self) -> Optional[str]:
        """
        Get first visible store ID (including hidden)

        :return: store id
        """
        for id in self.window.core.assistants.store.get_ids():
            if not self.window.core.assistants.store.is_hidden(id):
                return id
        return None

    def open_by_idx(self, idx: int):
        """
        Open editor by tab index

        :param idx: list index
        """
        store = self.window.core.assistants.store.get_by_idx(idx)
        if store is None:
            return
        self.current = store
        self.open(force=True)

    def update(self):
        """Update vector store editor"""
        self.reload_items()
        self.window.controller.assistant.editor.update_store_list()  # update stores list in assistant dialog

    def set_hide_thread(self, state: bool):
        """
        Toggle show thread stores

        :param state: state
        """
        self.window.core.config.set("assistant.store.hide_threads", state)
        self.update()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.15 01:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.config import BaseConfig, BasePlugin


class Config(BaseConfig):
    def __init__(self, plugin: BasePlugin = None, *args, **kwargs):
        super(Config, self).__init__(plugin)
        self.plugin = plugin

    def from_defaults(self, plugin: BasePlugin = None):
        """
        Set default options for plugin

        :param plugin: plugin instance
        """
        plugin.add_option(
            "provider",
            type="combo",
            value="google_custom_search",
            label="Provider",
            description="Select search engine provider, default: Google",
            tooltip="Select search engine provider",
            keys=plugin.get_provider_options(),
        )
        plugin.add_option(
            "num_pages",
            type="int",
            value=10,
            label="Number of pages to search",
            description="Number of max pages to search per query",
            min=1,
            max=None,
        )
        plugin.add_option(
            "max_open_urls",
            type="int",
            value=1,
            label="Number of max URLs to open at once",
            description="Number of max URLs to open at once",
            min=1,
            max=None,
        )
        plugin.add_option(
            "max_page_content_length",
            type="int",
            value=0,
            label="Max content characters",
            description="Max characters of page content to get (0 = unlimited)",
            min=0,
            max=None,
        )
        plugin.add_option(
            "chunk_size",
            type="int",
            value=20000,
            label="Per-page content chunk size",
            description="Per-page content chunk size (max characters per chunk)",
            min=1,
            max=None,
        )
        plugin.add_option(
            "raw",
            type="bool",
            value=True,
            label="Use raw content (without summarization)",
            description="Return raw content from web search instead of summarized content",
            tooltip="Use raw content for search",
        )
        plugin.add_option(
            "disable_ssl",
            type="bool",
            value=True,
            label="Disable SSL verify",
            description="Disables SSL verification when crawling web pages",
            tooltip="Disable SSL verify",
        )
        plugin.add_option(
            "img_thumbnail",
            type="bool",
            value=True,
            label="Show thumbnail images",
            description="Enable fetching thumbnails from opened websites",
            tooltip="Enable fetching thumbnails from opened websites",
        )
        plugin.add_option(
            "timeout",
            type="int",
            value=5,
            label="Timeout",
            description="Connection timeout (seconds)",
            tooltip="Connection timeout (seconds)",
        )
        plugin.add_option(
            "user_agent",
            type="text",
            value="Mozilla/5.0",
            label="User agent",
            description="User agent to use when making requests, default: Mozilla/5.0",
            tooltip="User agent to use when making requests",
        )
        plugin.add_option(
            "max_result_length",
            type="int",
            value=50000,
            label="Max result length",
            description="Max length of summarized result (characters)",
            min=0,
            max=None,
        )
        plugin.add_option(
            "summary_max_tokens",
            type="int",
            value=1500,
            label="Max summary tokens",
            description="Max tokens in output when generating summary",
            min=0,
            max=None,
        )
        plugin.add_option(
            "model_tmp_query",
            type="combo",
            value="gpt-4o-mini",
            label="Model for query in-memory index",
            description="Model used for query in-memory index for `web_index_query` command, "
                        "default: gpt-4o-mini",
            tooltip="Query model",
            use="models",
            tab="indexing",
        )
        plugin.add_option(
            "auto_index",
            type="bool",
            value=False,
            label="Auto-index all used URLs using Llama-index",
            description="If enabled, every URL used by the model will be automatically indexed using Llama-index "
                        "(persistent index)",
            tooltip="If enabled, every URL used by the model will be automatically indexed using Llama-index "
                    "(persistent index)",
            tab="indexing",
        )
        plugin.add_option(
            "idx",
            type="combo",
            value="base",
            label="Index to use",
            description="ID of index to use for web page indexing (persistent index)",
            tooltip="Index name",
            use="idx",
            tab="indexing",
        )
        plugin.add_option(
            "summary_model",
            type="combo",
            value="gpt-4o-mini",
            label="Model used for web page summarize",
            description="Model used for web page summarize, default: gpt-4o-mini",
            use="models",
            advanced=True,
        )
        plugin.add_option(
            "prompt_summarize",
            type="textarea",
            value="Summarize text in English in a maximum of 3 paragraphs, trying to find the most "
                  "important content that can help answer the following question: {query}",
            label="Summarize prompt",
            description="Prompt used for web search results summarize, use {query} "
                        "as a placeholder for search query",
            tooltip="Prompt",
            advanced=True,
        )
        plugin.add_option(
            "prompt_summarize_url",
            type="textarea",
            value="Summarize text in English in a maximum of 3 paragraphs, trying to find the most "
                  "important content.",
            label="Summarize prompt (URL open)",
            description="Prompt used for specified URL page summarize",
            tooltip="Prompt",
            advanced=True,
        )

        # commands
        plugin.add_cmd(
            "web_url_open",
            instruction="read and get text content from ANY website URL. Always open a max of {max_urls} URLs at a time.",
            params=[
                {
                    "name": "url",
                    "type": "str",
                    "description": "URL to website",
                    "required": True,
                },
            ],
            enabled=True,
            description="If enabled, model will be able to open URL and read text content from it",
        )
        plugin.add_cmd(
            "web_url_raw",
            instruction="read and get raw HTML body from ANY website URL. Always open a max of {max_urls} URLs at a time.",
            params=[
                {
                    "name": "url",
                    "type": "str",
                    "description": "URL to website",
                    "required": True,
                },
            ],
            enabled=True,
            description="If enabled, model will be able to open specified URL and get raw HTML body from it",
        )
        plugin.add_cmd(
            "web_search",
            instruction="search the Web for list of URLs, prepare search query itself, list of "
                        "URLs will be returned, 10 links per page max. After receiving the list of URLs, "
                        "choose the best matched URLs and use the `web_url_open` command to read the content. "
                        "Always open a max of {max_urls} URLs at a time.",
            params=[
                {
                    "name": "query",
                    "type": "str",
                    "description": "search query",
                    "required": True,
                },
                {
                    "name": "page",
                    "type": "int",
                    "description": "page number",
                    "required": False,
                },
                {
                    "name": "num_links",
                    "type": "int",
                    "description": "links per page",
                    "required": False,
                },
            ],
            enabled=True,
            description="If enabled, model will be able to search the Web and get founded URLs list",
        )
        plugin.add_cmd(
            "web_request",
            instruction="send HTTP request to specified URL or API endpoint, using `requests` library. Tip: to "
                        "send raw data, use `data` parameter. To send data in `application/x-www-form-urlencoded` format, "
                        "use `data_form` parameter. To send data in `application/json` format, use `data_json` parameter. "
                        "Sending data in `data_form` and `data_json` parameters will automatically set `Content-Type`. To "
                        "upload files, use `files` parameter. To send cookies, use `cookies` parameter. To set custom headers, "
                        "use `headers` parameter. You can combine all parameters.",
            params=[
                {
                    "name": "url",
                    "type": "str",
                    "description": "URL to website or API endpoint",
                    "required": True,
                },
                {
                    "name": "method",
                    "type": "str",
                    "description": "HTTP method, default: GET",
                    "required": False,
                },
                {
                    "name": "headers",
                    "type": "dict",
                    "description": "HTTP headers",
                    "required": False,
                },
                {
                    "name": "params",
                    "type": "dict",
                    "description": "GET parameters",
                    "required": False,
                },
                {
                    "name": "data",
                    "type": "str",
                    "description": "raw data to send in POST requests",
                    "required": False,
                },
                {
                    "name": "data_form",
                    "type": "dict",
                    "description": "data to send in POST `application/x-www-form-urlencoded` requests",
                    "required": False,
                },
                {
                    "name": "data_json",
                    "type": "dict",
                    "description": "JSON data to send in POST `application/json` requests",
                    "required": False,
                },
                {
                    "name": "cookies",
                    "type": "dict",
                    "description": "Cookies",
                    "required": False,
                },
                {
                    "name": "files",
                    "type": "dict",
                    "description": "Files to upload, key is form field name, value is absolute path to file",
                    "required": False,
                },
            ],
            enabled=True,
            description="If enabled, model will be able to send any HTTP request to specified URL or API endpoint",
        )
        plugin.add_cmd(
            "web_extract_links",
            instruction="open webpage and get list of all links from it",
            params=[
                {
                    "name": "url",
                    "type": "str",
                    "description": "URL to website",
                    "required": True,
                },
            ],
            enabled=True,
            description="If enabled, model will be able to open URL and get list of all links from it",
        )
        plugin.add_cmd(
            "web_extract_images",
            instruction="open webpage and get list of all images from it",
            params=[
                {
                    "name": "url",
                    "type": "str",
                    "description": "URL to website",
                    "required": True,
                },
                {
                    "name": "download",
                    "type": "bool",
                    "description": "Download images to disk if user wants to, default: False",
                    "required": False,
                },
            ],
            enabled=True,
            description="If enabled, model will be able to open URL and get list of all images from it",
        )
        plugin.add_cmd(
            "web_index",
            instruction="",
            params=[],  # prepared dynamically
            enabled=True,
            tab="indexing",
            description="If enabled, model will be able to index web content using Llama-index (persistent index)",
        )
        plugin.add_cmd(
            "web_index_query",
            instruction="read, index and query external content for additional context.",
            params=[],  # prepared dynamically
            enabled=True,
            tab="indexing",
            description="If enabled, model will be able to index and query web content using Llama-index "
                        "(temporary index, on the fly)",
        )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.05 23:00:00                  #
# ================================================== #


class Helpers:

    def __init__(self, window=None):
        """
        Helpers

        :param window: Window instance
        """
        self.window = window

    def pre_format_text(self, text: str) -> str:
        """
        Post-format text

        :param text: text to format
        :return: formatted text
        """
        text = text.strip()
        text = text.replace("#~###~", "~###~")  # fix for #~###~ in text (previous versions)
        text = text.replace("# ~###~", "~###~")  # fix for # ~###~ in text (previous versions)
        return text

    def post_format_text(self, text: str) -> str:
        """
        Post-format text

        :param text: text to format
        :return: formatted text
        """
        return text.strip()

    def format_user_text(self, text: str) -> str:
        """
        Post-format user text

        :param text: text to format
        :return: formatted text
        """
        return text

    def format_chunk(self, text: str) -> str:
        """
        Format chunk

        :param text: text to format
        :return: formatted text
        """
        return text



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.18 19:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.config import BaseConfig, BasePlugin


class Config(BaseConfig):
    def __init__(self, plugin: BasePlugin = None, *args, **kwargs):
        super(Config, self).__init__(plugin)
        self.plugin = plugin

    def from_defaults(self, plugin: BasePlugin = None):
        """
        Set default options for plugin

        :param plugin: plugin instance
        """
        keys = {
            "enabled": "bool",
            "crontab": "text",
            "prompt": "textarea",
            "preset": {
                "type": "combo",
                "use": "presets",
                "keys": [],
            },
        }
        items = [
            {
                "enabled": False,
                "crontab": "30 9 * * *",
                "prompt": "Hi! This prompt should be sent at 9:30 every day. What time is it?",
                "preset": "_",
            },
        ]
        desc = "Add your cron-style tasks here. They will be executed automatically at the times you specify in " \
               "the cron-based job format. If you are unfamiliar with Cron, consider visiting the Cron Guru " \
               "page for assistance."
        tooltip = "Check out the tutorials about Cron or visit the Crontab Guru for help on how to use Cron syntax."
        plugin.add_option(
            "crontab",
            type="dict",
            value=items,
            label="Your tasks",
            description=desc,
            tooltip=tooltip,
            keys=keys,
            urls={
                "Crontab Guru": "https://crontab.guru",
            },
        )
        plugin.add_option(
            "new_ctx",
            type="bool",
            value=True,
            label="Create a new context on job run",
            description="If enabled, then a new context will be created on every run of the job",
        )
        plugin.add_option(
            "show_notify",
            type="bool",
            value=True,
            label="Show notification on job run",
            description="If enabled, then a tray notification will be shown on every run of the job",
        )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.16 01:00:00                  #
# ================================================== #

import os.path
from typing import Optional

import chromadb
from chromadb.config import Settings

from llama_index.core.indices.base import BaseIndex
from llama_index.core import StorageContext
from llama_index.vector_stores.chroma import ChromaVectorStore

from .base import BaseStore


class ChromaProvider(BaseStore):
    def __init__(self, *args, **kwargs):
        super(ChromaProvider, self).__init__(*args, **kwargs)
        """
        Chroma vector store provider

        :param args: args
        :param kwargs: kwargs
        """
        self.window = kwargs.get('window', None)
        self.id = "ChromaVectorStore"
        self.prefix = "chroma_"  # prefix for index directory
        self.indexes = {}

    def get_db(self, id: str):
        """
        Get database instance

        :param id: index name
        :return: database instance
        """
        path = self.get_path(id)
        return chromadb.PersistentClient(
            path=path, 
            settings=Settings(
                anonymized_telemetry=False
            )
        )

    def create(
            self,
            id: str,
            embed_model: Optional = None):
        """
        Create empty index

        :param id: index name
        """
        path = self.get_path(id)
        if not os.path.exists(path):
            index = self.index_from_empty(embed_model)  # create empty index
            self.store(
                id=id,
                index=index,
            )

    def get(
            self,
            id: str,
            llm: Optional = None,
            embed_model: Optional = None,
    ) -> BaseIndex:
        """
        Get index

        :param id: index name
        :param llm: LLM instance
        :param embed_model: Embedding model instance
        :return: index instance
        """
        if not self.exists(id):
            self.create(id, embed_model)
        path = self.get_path(id)
        db = self.get_db(id)
        chroma_collection = db.get_or_create_collection(id)
        vector_store = ChromaVectorStore(
            chroma_collection=chroma_collection,
        )
        storage_context = StorageContext.from_defaults(
            persist_dir=path,
        )
        self.indexes[id] = self.index_from_store(
            vector_store=vector_store,
            storage_context=storage_context,
            llm=llm,
            embed_model=embed_model,
        )
        return self.indexes[id]

    def store(
            self,
            id: str,
            index: Optional[BaseIndex] = None
    ):
        """
        Store index

        :param id: index name
        :param index: index instance
        """
        if index is None:
            index = self.indexes[id]
        path = self.get_path(id)
        index.storage_context.persist(
            persist_dir=path,
        )
        self.indexes[id] = index



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.01.25 19:00:00                  #
# ================================================== #

from PySide6.QtCore import QRunnable, Slot, QObject, Signal


class Worker(QRunnable):
    def __init__(self, fn, *args, **kwargs):
        super(Worker, self).__init__()
        self.fn = fn
        self.args = args
        self.kwargs = kwargs

    @Slot()
    def run(self):
        self.fn(*self.args, **self.kwargs)


class WorkerSignals(QObject):
    updated = Signal(object)
    finished = Signal(object)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.19 02:00:00                  #
# ================================================== #

import copy
from typing import List, Dict

from PySide6 import QtCore
from PySide6.QtCore import QObject, QEvent

from pygpt_net.core.events import ControlEvent


class Shortcuts:
    def __init__(self, window=None):
        """
        Shortcuts core

        :param window: Window instance
        """
        self.window = window

    def get_keys_choices(self) -> List[Dict[str, str]]:
        """
        Get keys choices with identical key and value.

        :return: List of choices
        """
        keys = [{chr(i): chr(i)} for i in range(65, 91)]  # A-Z
        keys.extend([{str(i): str(i)} for i in range(0, 10)])  # 0-9
        keys.extend([{'F' + str(i): 'F' + str(i)} for i in range(1, 13)])  # F1-F12
        symbols = [
            'Space', 'Comma', 'Period', 'Slash', 'Backslash', 'Plus', 'Minus', 'Equal', 'BracketLeft', 'BracketRight',
            'Semicolon', 'Apostrophe', 'CapsLock', 'Up', 'Down', 'Left', 'Right', 'PageUp', 'PageDown', 'Home', 'End',
            'Insert', 'Delete', 'Enter', 'Return', 'Tab', 'Backspace', 'Escape', 'Print', 'SysReq', 'Pause', 'ScrollLock',
        ]
        keys.extend([{name: name} for name in symbols])
        return keys

    def get_modifiers_choices(self) -> List[Dict[str, str]]:
        """
        Get modifiers choices with identical key and value.

        :return: List of choices
        """
        # SHIFT not working properly
        modifiers_names = [
            '---', 'Ctrl', 'Alt', 'Shift',
        ]
        modifiers = [{name: name} for name in modifiers_names]
        return modifiers

    def handle_global_shortcuts(self, event: QEvent):
        """
        Handle global shortcuts

        :param event: event
        """
        # esc key
        if event.type() == QEvent.KeyPress and event.key() == QtCore.Qt.Key_Escape:
            self.window.controller.access.on_escape()
            return True

        if not hasattr(self.window, 'core') or not hasattr(self.window.core, 'config'):
            return False

        if event.type() == QEvent.KeyPress:
            # shortcuts
            config = copy.deepcopy(self.window.core.config.get("access.shortcuts"))
            for shortcut in config:
                if shortcut['key'] == "" or shortcut['key'] is None:
                    continue
                key_name = 'Key_' + str(shortcut['key'])
                if hasattr(QtCore.Qt, key_name):
                    shortcut['key'] = getattr(QtCore.Qt, 'Key_' + str(shortcut['key']))
                else:
                    continue
                modifier_name = str(shortcut['key_modifier']) + 'Modifier'
                if shortcut['key_modifier'] == "" or shortcut['key_modifier'] is None or shortcut[
                    'key_modifier'] == "---":
                    shortcut['key_modifier'] = QtCore.Qt.NoModifier
                elif hasattr(QtCore.Qt, modifier_name):
                    shortcut['key_modifier'] = getattr(QtCore.Qt, modifier_name, QtCore.Qt.NoModifier)

            for shortcut in config:
                if (event.key() == shortcut['key'] and
                        (shortcut['key_modifier'] == QtCore.Qt.NoModifier or event.modifiers() == shortcut[
                            'key_modifier'])
                ):
                    self.window.dispatch(ControlEvent(shortcut['action']))
                    return True
        return False


class GlobalShortcutFilter(QObject):
    def __init__(self, window=None):
        super().__init__()
        self.window = window

    def eventFilter(self, obj, event):
        """
        Event filter for global shortcuts handling

        :param obj: object
        :param event: event
        :return: True if event was handled
        """
        try:
            if event.type() != QEvent.KeyPress:
                return False
            if not self.window or not hasattr(self.window, 'controller'):
                return False
            return self.window.core.access.shortcuts.handle_global_shortcuts(event)
        except Exception as e:
            print(e)
        return False


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.15 01:00:00                  #
# ================================================== #

import ssl
from urllib.request import Request, urlopen

from pygpt_net.plugin.base.plugin import BasePlugin
from pygpt_net.provider.web.base import BaseProvider
from pygpt_net.core.events import Event
from pygpt_net.item.ctx import CtxItem

from .config import Config
from .websearch import WebSearch
from .worker import Worker


class Plugin(BasePlugin):
    def __init__(self, *args, **kwargs):
        super(Plugin, self).__init__(*args, **kwargs)
        self.id = "cmd_web"
        self.name = "Web Search"
        self.description = "Allows to connect to the Web and search web pages for actual data."
        self.prefix = "Web"
        self.urls = {}
        self.input_text = None
        self.allowed_cmds = [
            "web_search",
            "web_urls",
            "web_url_open",
            "web_url_raw",
            "web_request",
            "web_index",
            "web_index_query",
            "web_extract_links",
            "web_extract_images",
        ]
        self.order = 100
        self.use_locale = True
        self.worker = None
        self.websearch = WebSearch(self)
        self.config = Config(self)

    def init_options(self):
        """Initialize options"""
        self.config.from_defaults(self)
        self.init_provider()
        self.init_idx_options()

    def init_provider(self):
        """Initialize provider options"""
        providers = self.get_providers()
        for id in providers:
            providers[id].init(self)

    def init_idx_options(self):
        """Initialize indexing options"""
        instructions = self.window.core.idx.indexing.get_external_instructions()
        providers = self.window.core.idx.indexing.get_data_providers()
        # enable/disable indexing for each loader
        for id in list(instructions.keys()):
            name = id
            if id in providers:
                name = providers[id].name
            self.add_option(
                "index_web_" + id,
                type="bool",
                value=True,
                label="Enable: " + name,
                description="Enable indexing data from web loader: " + id,
                advanced=False,
                locale=False,
                tab="indexing",
            )

    def is_indexing_allowed(self, id: str) -> bool:
        """
        Check if indexing from specified loader is allowed

        :param id: loader ID
        :return: True if allowed
        """
        return self.get_option_value("index_web_" + id)

    def get_providers(self) -> dict:
        """
        Get web search providers

        :return: providers dict
        """
        return self.window.core.web.get_providers(self.window.core.web.PROVIDER_SEARCH_ENGINE)

    def get_provider_options(self) -> list:
        """Get provider options"""
        options = []
        providers = self.get_providers()
        for id in providers:
            options.append({id: providers[id].name})
        return options

    def init_tabs(self) -> dict:
        """
        Initialize provider tabs

        :return: dict of tabs
        """
        tabs = {}
        tabs["general"] = "General"
        tabs["indexing"] = "Indexing"
        providers = self.get_providers()
        for id in providers:
            tabs[id] = providers[id].name
        return tabs

    def attach(self, window):
        """
        Attach window

        :param window: Window
        """
        self.window = window

        # register provider tabs
        self.tabs = self.init_tabs()

        # register options
        self.init_options()

    def get_provider(self) -> BaseProvider:
        """
        Get search engine provider

        :return: provider instance
        """
        current = self.get_option_value("provider")
        providers = self.get_providers()
        if current not in providers:
            raise Exception("Provider '{}' not found!".format(current))
        return providers[current]

    def handle(self, event: Event, *args, **kwargs):
        """
        Handle dispatched event

        :param event: event object
        :param args: event args
        :param kwargs: event kwargs
        """
        name = event.name
        data = event.data
        ctx = event.ctx

        if name == Event.INPUT_BEFORE:
            self.on_input_before(data['value'])

        elif name == Event.CMD_SYNTAX:
            self.cmd_syntax(data)

        elif name == Event.CMD_EXECUTE:
            self.cmd(ctx, data['commands'])

    def on_input_before(self, text: str):
        """
        Event: INPUT_BEFORE

        :param text: input text
        """
        self.input_text = text

    def prepare_idx_syntax(self) -> dict:
        """
        Prepare web_index command syntax with instructions parsed from web loaders

        :return: syntax string
        """
        instructions = self.window.core.idx.indexing.get_external_instructions()
        types = {}
        args = {}
        for type in instructions:
            if not self.is_indexing_allowed(type):  # check if data loader is allowed
                continue
            types[type] = instructions[type]["description"]
            args[type] = list(instructions[type]["args"].keys())

        cmd = {
            "cmd": "web_index",
            "instruction": "read and index content from external source. Use it to read and index as vectors "
                        "(embed in vector DB) the provided URL for webpage or any other remote resource, like "
                        "YT video, RSS, etc. Provide type of the resource in the \"type\" param. If there is no allowed "
                        "type for a specified resource then use the default \"webpage\" type. If selected type requires "
                        "additional args, then pass them into \"args\"",
            "params": [
                {
                    "name": "type",
                    "type": "enum",
                    "description": "data type",
                    "required": True,
                    "default": "webpage",
                    "enum": {
                        "type": types,
                    }
                },
                {
                    "name": "args",
                    "type": "dict",
                    "description": "extra args for type",
                    "required": True,
                    "enum": {
                        "args": args,
                    }
                }
            ],
            "enabled": True,
        }
        return cmd

    def prepare_idx_query_syntax(self) -> dict:
        """
        Prepare web_index command syntax with instructions parsed from web loaders

        :return: syntax string
        """
        instructions = self.window.core.idx.indexing.get_external_instructions()
        types = {}
        args = {}
        for type in instructions:
            if not self.is_indexing_allowed(type):  # check if data loader is allowed
                continue
            types[type] = instructions[type]["description"]
            args[type] = list(instructions[type]["args"].keys())

        cmd = {
            "cmd": "web_index_query",
            "instruction": "read, index and query external content for additional context",
            "params": [
                {
                    "name": "type",
                    "type": "enum",
                    "description": "data type",
                    "required": True,
                    "default": "webpage",
                    "enum": {
                        "type": types,
                    }
                },
                {
                    "name": "args",
                    "type": "dict",
                    "description": "extra args for type",
                    "required": True,
                    "enum": {
                        "args": args,
                    }
                }
            ],
            "enabled": True,
        }
        return cmd

    def cmd_syntax(self, data: dict):
        """
        Event: CMD_SYNTAX

        :param data: event data dict
        """
        max_urls = self.get_option_value("max_open_urls")
        is_summary = True
        if self.get_option_value("raw"):
            is_summary = False

        for option in self.allowed_cmds:
            if not self.has_cmd(option):
                continue

            if option == "web_index":
                data['cmd'].append(self.prepare_idx_syntax())  # prepare params
                continue
            elif option == "web_index_query":
                data['cmd'].append(self.prepare_idx_query_syntax())  # prepare params
                continue
            elif option == "web_url_open":
                cmd = self.get_cmd(option)
                try:
                    cmd["instruction"] = cmd["instruction"].format(
                        max_urls=max_urls,
                    )
                    if not is_summary:
                        for param in list(cmd["params"]):
                            if param["name"] == "summarize_prompt":
                                cmd["params"].remove(param)
                    data['cmd'].append(cmd)
                    continue
                except Exception as e:
                    pass
            elif option == "web_url_raw":
                cmd = self.get_cmd(option)
                try:
                    cmd["instruction"] = cmd["instruction"].format(
                        max_urls=max_urls,
                    )
                    if not is_summary:
                        for param in list(cmd["params"]):
                            if param["name"] == "summarize_prompt":
                                cmd["params"].remove(param)
                    data['cmd'].append(cmd)
                    continue
                except Exception as e:
                    pass
            elif option == "web_search":
                max_pages = self.get_option_value("num_pages")
                cmd = self.get_cmd(option)
                try:
                    cmd["instruction"] = cmd["instruction"].format(
                        max_pages=max_pages,
                        max_urls=max_urls,
                    )
                    if not is_summary:
                        for param in list(cmd["params"]):
                            if param["name"] == "summarize_prompt":
                                cmd["params"].remove(param)
                    data['cmd'].append(cmd)
                    continue
                except Exception as e:
                    pass

            data['cmd'].append(self.get_cmd(option))  # append command

    def cmd(self, ctx: CtxItem, cmds: list):
        """
        Event: CMD_EXECUTE

        :param ctx: CtxItem
        :param cmds: commands dict
        """
        my_commands = []
        for item in cmds:
            if item["cmd"] in self.allowed_cmds and self.has_cmd(item["cmd"]):
                my_commands.append(item)

        if len(my_commands) == 0:
            return

        # check required API keys and other settings
        if not self.get_provider().is_configured(my_commands):
            self.window.ui.dialogs.alert(self.get_provider().get_config_message())
            self.gen_api_key_response(ctx, cmds)
            return

        # set state: busy
        self.cmd_prepare(ctx, my_commands)

        try:
            worker = Worker()
            worker.from_defaults(self)
            worker.cmds = my_commands
            worker.ctx = ctx
            worker.websearch = self.websearch
            if not self.is_async(ctx):
                worker.run()
                return
            worker.run_async()

        except Exception as e:
            self.error(e)

    def gen_api_key_response(self, ctx: CtxItem, cmds: list):
        """
        Generate response for empty API key error

        :param ctx: CtxItem
        :param cmds: commands dict
        """
        for item in cmds:
            request = {
                "cmd": item["cmd"],
            }
            err = "API key is not set. Please set credentials in plugin settings."
            self.log(err)
            self.window.update_status(err)
            msg = "Tell the user that the required API key or other config is not configured in the plugin settings, " \
                  "and to set the API key in the settings in order to use the internet search plugin."
            data = {
                'msg_to_user': msg,
            }
            response = {
                "request": request,
                "result": data,
            }
            self.reply(response, ctx)
            return

    def get_url(self, url: str, extra_headers: dict = None) -> bytes:
        """
        Get URL content

        :param url: URL
        :param extra_headers: extra headers
        :return: response data (bytes)
        """
        headers = {
            'User-Agent': self.get_option_value("user_agent"),
        }
        if extra_headers is not None:
            headers.update(extra_headers)

        req = Request(
            url=url,
            headers=headers,
        )
        data = ""
        try:
            if self.get_option_value('disable_ssl'):
                context = ssl.create_default_context()
                context.check_hostname = False
                context.verify_mode = ssl.CERT_NONE
                data = urlopen(
                    req,
                    context=context,
                    timeout=self.get_option_value('timeout'),
                ).read()
            else:
                data = urlopen(
                    req,
                    timeout=self.get_option_value('timeout'),
                ).read()
        except Exception as e:
            data = str(e)
        return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 00:00:00                  #
# ================================================== #

from typing import Any

from pygpt_net.core.events import AppEvent
from pygpt_net.item.ctx import CtxItem

from .attachment import Attachment
from .audio import Audio
from .command import Command
from .common import Common
from .files import Files
from .image import Image
from .input import Input
from .output import Output
from .render import Render
from .response import Response
from .stream import Stream
from .text import Text
from .vision import Vision


class Chat:
    def __init__(self, window=None):
        """
        Chat controller

        :param window: Window instance
        """
        self.window = window
        self.attachment = Attachment(window)
        self.audio = Audio(window)
        self.command = Command(window)
        self.common = Common(window)
        self.files = Files(window)
        self.image = Image(window)
        self.input = Input(window)
        self.output = Output(window)
        self.render = Render(window)
        self.response = Response(window)
        self.stream = Stream(window)
        self.text = Text(window)
        self.vision = Vision(window)

    def init(self):
        """Init"""
        self.render.setup()  # setup render engine

    def setup(self):
        """Setup"""
        self.common.setup()
        self.attachment.setup()

    def reload(self):
        """Reload"""
        self.common.setup()
        self.render.reload()
        self.attachment.reload()

    def handle_error(self, err: Any):
        """
        Handle error

        :param err: Exception
        """
        self.window.core.debug.log(err)
        self.window.ui.dialogs.alert(str(err))
        self.window.update_status(str(err))
        self.window.controller.chat.common.unlock_input()  # always unlock input on error
        self.window.stateChanged.emit(self.window.STATE_ERROR)
        self.window.dispatch(AppEvent(AppEvent.INPUT_ERROR))  # app event

        # stop agent on error
        if self.window.controller.agent.legacy.enabled():
            self.window.controller.agent.legacy.on_stop()

    def log_ctx(self, ctx: CtxItem, mode: str):
        """
        Log context item

        :param ctx: CtxItem
        :param mode: mode (input/output)
        """
        if self.window.core.config.get("log.ctx"):
            self.log("[ctx] {}: {}".format(mode.upper(), ctx.dump()))  # log
        else:
            self.log("[ctx] {}.".format(mode.upper()))

    def log(self, data: Any):
        """
        Log data to debug

        :param data: Data to log
        """
        self.window.core.debug.info("[chat] " + str(data))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 19:00:00                  #
# ================================================== #

from typing import Optional, List, Dict

from pygpt_net.provider.web.base import BaseProvider

from .helpers import Helpers


class Web:

    PROVIDER_SEARCH_ENGINE = "search_engine"

    def __init__(self, window=None):
        """
        Web core

        :param window: Window instance
        """
        self.window = window
        self.helpers = Helpers(window)
        self.providers = {
            self.PROVIDER_SEARCH_ENGINE: {},
        }

    def is_registered(
            self,
            id: str,
            type: str = PROVIDER_SEARCH_ENGINE
    ) -> bool:
        """
        Check if provider is registered

        :param id: provider id
        :param type: provider type
        :return: True if registered
        """
        if type in self.providers:
            return id in self.providers[type]
        return False

    def get_providers(
            self,
            type: str = PROVIDER_SEARCH_ENGINE
    ) -> Dict[str, BaseProvider]:
        """
        Get all providers

        :param type: provider type
        :return: providers dict
        """
        if type in self.providers:
            return self.providers[type]
        return {}

    def get_ids(
            self,
            type: str = PROVIDER_SEARCH_ENGINE
    ) -> List[str]:
        """
        Get all providers ids

        :param type: provider type
        :return: providers ids list
        """
        if type in self.providers:
            return list(self.providers[type].keys())
        return []

    def get(
            self,
            id: str,
            type: str = PROVIDER_SEARCH_ENGINE
    ) -> Optional[BaseProvider]:
        """
        Get provider instance

        :param id: provider id
        :param type: provider type
        :return: provider instance
        """
        if self.is_registered(id, type):
            return self.providers[type][id]
        return None

    def register(self, provider: BaseProvider):
        """
        Register provider

        :param provider: provider instance
        """
        id = provider.id
        type = provider.type
        for t in type:
            if t in self.providers:
                self.providers[t][id] = provider



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.16 01:00:00                  #
# ================================================== #

import os

from pygpt_net.plugin.base.plugin import BasePlugin
from pygpt_net.core.events import Event
from pygpt_net.item.ctx import CtxItem

from .config import Config
from .output import Output
from .worker import Worker


class Plugin(BasePlugin):
    def __init__(self, *args, **kwargs):
        super(Plugin, self).__init__(*args, **kwargs)
        self.id = "cmd_files"
        self.name = "Files I/O"
        self.description = "Provides commands to read and write files"
        self.prefix = "I/O"
        self.order = 100
        self.allowed_cmds = [
            "read_file",
            "query_file",
            "save_file",
            "append_file",
            "delete_file",
            "list_dir",
            "mkdir",
            "download_file",
            "rmdir",
            "copy_file",
            "copy_dir",
            "move",
            "find",
            "is_dir",
            "is_file",
            "tree",
            "file_exists",
            "file_size",
            "file_info",
            "send_file",
            "cwd",
            "file_index",
        ]
        self.use_locale = True
        self.worker = None
        self.output = Output(self)
        self.config = Config(self)
        self.init_options()

    def init_options(self):
        """Initialize options"""
        self.config.from_defaults(self)

    def handle(self, event: Event, *args, **kwargs):
        """
        Handle dispatched event

        :param event: event object
        :param args: event args
        :param kwargs: event kwargs
        """
        name = event.name
        data = event.data
        ctx = event.ctx

        if name == Event.CMD_SYNTAX:
            self.cmd_syntax(data)

        elif name == Event.CMD_EXECUTE:
            self.cmd(
                ctx,
                data['commands'],
            )

        elif name == Event.TOOL_OUTPUT_RENDER:
            if data['tool'] == self.id:
                data['html'] = self.output.handle(ctx, data['content'])

    def cmd_syntax(self, data: dict):
        """
        Event: CMD_SYNTAX

        :param data: event data dict
        """
        for option in self.allowed_cmds:
            if self.has_cmd(option):
                data['cmd'].append(self.get_cmd(option))  # append command

    def cmd(self, ctx: CtxItem, cmds: list):
        """
        Event: CMD_EXECUTE

        :param ctx: CtxItem
        :param cmds: commands dict
        """
        is_cmd = False
        my_commands = []
        for item in cmds:
            if item["cmd"] in self.allowed_cmds:
                my_commands.append(item)
                is_cmd = True

        if not is_cmd:
            return

        # set state: busy
        self.cmd_prepare(ctx, my_commands)

        try:
            worker = Worker()
            worker.from_defaults(self)
            worker.cmds = my_commands
            worker.ctx = ctx

            if not self.is_async(ctx):
                worker.run()
                return
            worker.run_async()

        except Exception as e:
            self.error(e)

    def read_as_text(self, path: str, use_loaders: bool = True) -> str:
        """
        Read file and return content as text

        :param path: file path
        :param use_loaders: use Llama-index loader to read file
        :return: text content
        """
        # use_loaders = False
        if use_loaders:
            content, docs = str(self.window.core.idx.indexing.read_text_content(path))
            return content
        else:
            data = ""
            if os.path.isfile(path):
                with open(path, 'r', encoding="utf-8") as file:
                    data = file.read()
            return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.19 02:00:00                  #
# ================================================== #

from pygpt_net.ui.widget.tabs.body import TabBody
from pygpt_net.ui.widget.filesystem.explorer import FileExplorer


class Explorer:
    def __init__(self, window=None):
        """
        Chat output UI

        :param window: Window instance
        """
        self.window = window

    def setup(self) -> TabBody:
        """
        Setup file explorer

        :return: TabBody
        """
        # index status data
        index_data = self.window.core.idx.get_idx_data()  # get all idx data

        # file explorer
        path = self.window.core.config.get_user_dir('data')
        self.window.ui.nodes['output_files'] = FileExplorer(self.window, path, index_data)

        # build tab body
        body = self.window.core.tabs.from_widget(self.window.ui.nodes['output_files'])
        body.append(self.window.ui.nodes['output_files'])
        return body



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.15 00:00:00                  #
# ================================================== #

import os


class ModelsDebug:
    def __init__(self, window=None):
        """
        Models debug

        :param window: Window instance
        """
        self.window = window
        self.id = 'models'

    def update(self):
        """Update debug window."""
        self.window.core.debug.begin(self.id)

        path = os.path.join(self.window.core.config.path, '', 'models.json')
        self.window.core.debug.add(self.id, 'Models File', str(path))

        self.window.core.debug.add(
            self.id, 'Options',
            str(self.window.controller.model.editor.get_options())
        )

        # models
        for key in self.window.core.models.items:
            if key == '__meta__':
                self.window.core.debug.add(self.id, '__meta__', str(self.window.core.models.items[key]))
                continue
            model = self.window.core.models.items[key]
            data = {
                'id': model.id,
                'name': model.name,
                'mode': model.mode,
                'multimodal': model.multimodal,
                'langchain': model.langchain,
                'llama_index': model.llama_index,
                'tokens': model.tokens,
                'ctx': model.ctx,
                'default': model.default,
            }
            self.window.core.debug.add(self.id, str(key), str(data))

        self.window.core.debug.end(self.id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

import os
from typing import Optional, List, Dict

from pygpt_net.core.events import Event
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans

from .syntax_highlight import SyntaxHighlight

import pygpt_net.js_rc
import pygpt_net.css_rc
import pygpt_net.fonts_rc

class Body:

    def __init__(self, window=None):
        """
        HTML Body

        :param window: Window instance
        """
        self.window = window
        self.highlight = SyntaxHighlight(window)

    def is_timestamp_enabled(self) -> bool:
        """
        Check if timestamp is enabled

        :return: True if timestamp is enabled
        """
        return self.window.core.config.get('output_timestamp')

    def prepare_styles(self) -> str:
        """
        Prepare CSS styles

        :return: CSS styles
        """
        syntax_dark = [
            "dracula",
            "fruity",
            "github-dark",
            "gruvbox-dark",
            "inkpot",
            "material",
            "monokai",
            "native",
            "nord",
            "nord-darker",
            "one-dark",
            "paraiso-dark",
            "rrt",
            "solarized-dark",
            "stata-dark",
            "vim",
            "zenburn",
        ]
        syntax_style = self.window.core.config.get("render.code_syntax")
        if syntax_style is None or syntax_style == "":
            syntax_style = "default"
        fonts_path = os.path.join(self.window.core.config.get_app_path(), "data", "fonts").replace("\\", "/")
        stylesheet = self.window.controller.theme.markdown.get_web_css().replace('%fonts%', fonts_path)
        if syntax_style in syntax_dark:
            stylesheet += "pre { color: #fff; }"
        else:
            stylesheet += "pre { color: #000; }"
        return stylesheet + " " + self.highlight.get_style_defs()

    def prepare_action_icons(
            self,
            ctx: CtxItem
    ) -> str:
        """
        Append action icons

        :param ctx: context item
        :return: HTML code
        """
        html = ""
        show_edit = True
        # show_edit = self.window.core.config.get('ctx.edit_icons')
        icons_html = "".join(self.get_action_icons(ctx, all=show_edit))
        if icons_html != "":
            extra = "<div class=\"action-icons\" data-id=\"{}\">{}</div>".format(ctx.id, icons_html)
            html += extra
        return html

    def get_action_icons(
            self,
            ctx: CtxItem,
            all: bool = False
    ) -> List[str]:
        """
        Get action icons for context item

        :param ctx: context item
        :param all: True to show all icons
        :return: list of icons
        """
        icons = []

        # audio read
        if ctx.output is not None and ctx.output != "":
            icons.append(
                '<a href="extra-audio-read:{}" class="action-icon" data-id="{}" role="button"><span class="cmd">{}</span></a>'.format(
                    ctx.id,
                    ctx.id,
                    self.get_icon("audio", trans("ctx.extra.audio"), ctx)
                )
            )
            # copy ctx
            icons.append(
                '<a href="extra-copy:{}" class="action-icon" data-id="{}" role="button"><span class="cmd">{}</span></a>'.format(
                    ctx.id,
                    ctx.id,
                    self.get_icon("copy", trans("ctx.extra.copy"), ctx)
                )
            )
            # regen link
            icons.append(
                '<a href="extra-replay:{}" class="action-icon" data-id="{}" role="button"><span class="cmd">{}</span></a>'.format(
                    ctx.id,
                    ctx.id,
                    self.get_icon("reload", trans("ctx.extra.reply"), ctx)
                )
            )
            # edit link
            icons.append(
                '<a href="extra-edit:{}" class="action-icon edit-icon" data-id="{}" role="button"><span class="cmd">{}</span></a>'.format(
                    ctx.id,
                    ctx.id,
                    self.get_icon("edit", trans("ctx.extra.edit"), ctx)
                )
            )
            # delete link
            icons.append(
                '<a href="extra-delete:{}" class="action-icon edit-icon" data-id="{}" role="button"><span class="cmd">{}</span></a>'.format(
                    ctx.id,
                    ctx.id,
                    self.get_icon("delete", trans("ctx.extra.delete"), ctx)
                )
            )

            # join link
            if not self.window.core.ctx.is_first_item(ctx.id):
                icons.append(
                    '<a href="extra-join:{}" class="action-icon edit-icon" data-id="{}" role="button"><span class="cmd">{}</span></a>'.format(
                        ctx.id,
                        ctx.id,
                        self.get_icon("join", trans("ctx.extra.join"), ctx)
                    )
                )
        return icons

    def get_icon(
            self,
            icon: str,
            title: Optional[str] = None,
            item: Optional[CtxItem] = None
    ) -> str:
        """
        Get icon

        :param icon: icon name
        :param title: icon title
        :param item: context item
        :return: icon HTML
        """
        icon = os.path.join(self.window.core.config.get_app_path(), "data", "icons", "chat", icon + ".png")
        return '<img src="file://{}" class="action-img" title="{}" alt="{}" data-id="{}">'.format(
            icon, title, title, item.id)

    def get_image_html(
            self,
            url: str,
            num: Optional[int] = None,
            num_all: Optional[int] = None
    ) -> str:
        """
        Get image HTML

        :param url: URL to image
        :param num: number of image
        :param num_all: number of all images
        :return: HTML code
        """
        num_str = ""
        if num is not None and num_all is not None and num_all > 1:
            num_str = " [{}]".format(num)
        url, path = self.window.core.filesystem.extract_local_url(url)
        basename = os.path.basename(path)
        return """<div class="extra-src-img-box" title="{url}"><div class="img-outer"><div class="img-wrapper"><a href="{url}"><img src="{path}" class="image"></a></div>
        <a href="{url}" class="title">{title}</a></div></div>""". \
            format(prefix=trans('chat.prefix.img'),
                   url=url,
                   title=basename,
                   path=path,
                   num=num_str)

    def get_url_html(
            self,
            url: str,
            num: Optional[int] = None,
            num_all: Optional[int] = None
    ) -> str:
        """
        Get URL HTML

        :param url: external URL
        :param num: number of URL
        :param num_all: number of all URLs
        :return: HTML code
        """
        icon_path = os.path.join(
            self.window.core.config.get_app_path(),
            "data", "icons", "language.svg"
        )
        icon = '<img src="file://{}" class="extra-src-icon">'.format(icon_path)
        num_str = ""
        if num is not None and num_all is not None and num_all > 1:
            num_str = " [{}]".format(num)
        return """{icon}<a href="{url}" title="{url}">{url}</a> <small>{num}</small>""". \
            format(url=url,
                   num=num_str,
                   icon=icon,
            )

    def get_docs_html(self, docs: List[Dict]) -> str:
        """
        Get Llama-index doc metadata HTML

        :param docs: list of document metadata
        :return: HTML code
        """
        html = ""
        html_sources = ""
        num = 1
        max = 3
        try:
            for item in docs:
                for uuid, doc_json in item.items():
                    """
                    Example doc (file metadata):
                    {'a2c7af6d-3c34-4c28-bf2d-6161e7fb525e': {
                        'file_path': '/home/user/.config/pygpt-net/data/my_cars.txt',
                        'file_name': '/home/user/.config/pygpt-net/data/my_cars.txt', 'file_type': 'text/plain',
                        'file_size': 28, 'creation_date': '2024-03-03', 'last_modified_date': '2024-03-03',
                        'last_accessed_date': '2024-03-03'}}
                    """
                    doc_parts = []
                    for key in doc_json:
                        doc_parts.append("<b>{}:</b> {}".format(key, doc_json[key]))
                    html_sources += "<p><small>[{}] {}: {}</small></p>".format(num, uuid, ", ".join(doc_parts))
                    num += 1
                if num >= max:
                    break
        except Exception as e:
            pass

        icon_path = os.path.join(
            self.window.core.config.get_app_path(),
            "data", "icons", "db.svg"
        )
        icon = '<img src="file://{}" class="extra-src-icon">'.format(icon_path)
        if html_sources != "":
            html += "<p>{icon}<small><b>{prefix}:</b></small></p>".format(
                prefix=trans('chat.prefix.doc'),
                icon=icon,
            )
            html += "<div class=\"cmd\">"
            html += "<p>" + html_sources + "</p>"
            html += "</div> "
        return html

    def get_file_html(
            self,
            url: str,
            num: Optional[int] = None,
            num_all: Optional[int] = None
    ) -> str:
        """
        Get file HTML

        :param url: URL to file
        :param num: number of file
        :param num_all: number of all files
        :return: HTML code
        """
        icon_path = os.path.join(
            self.window.core.config.get_app_path(),
            "data", "icons", "attachments.svg"
        )
        icon = '<img src="file://{}" class="extra-src-icon">'.format(icon_path)
        num_str = ""
        if num is not None and num_all is not None and num_all > 1:
            num_str = " [{}]".format(num)
        url, path = self.window.core.filesystem.extract_local_url(url)
        return """{icon} <b>{num}</b> <a href="{url}">{path}</a>""". \
            format(url=url,
                   path=path,
                   num=num_str,
                   icon=icon,
            )

    def prepare_tool_extra(self, ctx: CtxItem) -> str:
        """
        Prepare footer extra

        :param ctx: context item
        :return: HTML code
        """
        html = ""
        if ctx.extra is not None and ctx.extra != "":
            html += "<div class=\"msg-extra\">"

            # single tool
            if "plugin" in ctx.extra:
                event = Event(Event.TOOL_OUTPUT_RENDER, {
                    'tool': ctx.extra["plugin"],
                    'html': '',
                    'multiple': False,
                    'content': ctx.extra,  # tool output
                })
                event.ctx = ctx
                self.window.dispatch(event, all=True)  # handle by plugins
                html += "<div class=\"tool-output-block\">" + event.data['html'] + "</div>"

            # multiple tools, list
            elif "tool_output" in ctx.extra and isinstance(ctx.extra["tool_output"], list):
                for key, tool in enumerate(ctx.extra["tool_output"]):
                    if "plugin" not in tool:
                        continue
                    event = Event(Event.TOOL_OUTPUT_RENDER, {
                        'tool': tool["plugin"],
                        'html': '',
                        'multiple': True,
                        'content': tool,  # tool output[]
                    })
                    event.ctx = ctx
                    self.window.dispatch(event, all=True)
                    html += "<div class=\"tool-output-block\">" + event.data['html'] + "</div>"
            html += "</div>"
        return html

    def get_html(self, pid: int) -> str:
        """
        Build webview HTML code

        :return: HTML code
        """
        classes = []
        classes_str = ""
        style = self.window.core.config.get("theme.style", "blocks")
        if self.window.core.config.get('render.blocks'):
            classes.append("display-blocks")
        if self.window.core.config.get('ctx.edit_icons'):
            classes.append("display-edit-icons")
        if self.is_timestamp_enabled():
            classes.append("display-timestamp")
        classes.append("theme-" + style)
        if classes:
            classes_str = ' class="' + " ".join(classes) + '"'

        return """
        <!DOCTYPE html>
        <html>
        <head>
            <style>
                """ + self.prepare_styles() + """
            </style>
        </head>
        <body """ + classes_str + """>
        <div id="container">
            <div id="_nodes_" class="nodes empty_list"></div>
            <div id="_append_input_" class="append_input"></div>
            <div id="_append_output_" class="append_output"></div>
            <div id="_footer_" class="footer"></div>
        </div>
        
        <link rel="stylesheet" href="qrc:///css/katex.min.css">
        <script type="text/javascript" src="qrc:///qtwebchannel/qwebchannel.js"></script>
        <script type="text/javascript" src="qrc:///js/highlight.min.js"></script>
        <script type="text/javascript" src="qrc:///js/katex.min.js"></script>
        <script type="text/javascript" src="qrc:///js/auto-render.min.js"></script>
        <script>
        
        let scrollTimeout = null;
        let prevScroll = 0;
        let bridge;
        let pid = """ + str(pid) + """
        new QWebChannel(qt.webChannelTransport, function (channel) {
            bridge = channel.objects.bridge;
        });
        history.scrollRestoration = "manual";
        document.addEventListener('keydown', function(event) {
            if (event.ctrlKey && event.key === 'f') {
                window.location.href = 'bridge://open_find:' + pid; // send to bridge
                event.preventDefault();
            }
            if (event.key === 'Escape') {
                window.location.href = 'bridge://escape'; // send to bridge
                event.preventDefault();
            }
        });
        document.addEventListener('click', function(event) {
            if (event.target.tagName !== 'A' && !event.target.closest('a')) {
                window.location.href = 'bridge://focus';
            }
        });
        function highlightCode() {
            document.querySelectorAll('pre code').forEach(el => {
                if (!el.classList.contains('hljs')) hljs.highlightElement(el);
            });
            renderMath();            
        }
        function renderMath() {
              const scripts = document.querySelectorAll('script[type^="math/tex"]');
              scripts.forEach(function(script) {
                const displayMode = script.type.indexOf('mode=display') > -1;
                const mathContent = script.textContent || script.innerText;
                const element = document.createElement(displayMode ? 'div' : 'span');
                try {
                  katex.render(mathContent, element, {
                    displayMode: displayMode,
                    throwOnError: false
                  });
                } catch (err) {
                  element.textContent = mathContent;
                }
                script.parentNode.replaceChild(element, script);
              });
        }
        function scrollToBottom() {
            getScrollPosition();  // store using bridge
            if (scrollTimeout !== null) {
                return;
            }
            if (document.body.scrollHeight > prevScroll) {
                scrollTimeout = setTimeout(function() {
                    window.scrollTo(0, document.body.scrollHeight);
                    prevScroll = document.body.scrollHeight;
                    getScrollPosition();  // store using bridge
                    scrollTimeout = null;
                }, 30);
            }
        }
        function appendToInput(content) {
            const element = document.getElementById('_append_input_');
            if (element) {
                element.innerHTML += content;
            }
            highlightCode();
            scrollToBottom();
        }
        function appendToOutput(bot_name, content) {
            const element = document.getElementById('_append_output_');
            if (element) {
                let box = element.querySelector('.msg-box');
                let msg;
                if (!box) {
                    box = document.createElement('div');
                    box.classList.add('msg-box');
                    box.classList.add('msg-bot');
                    var name = document.createElement('div');
                    name.classList.add('name-header');
                    name.classList.add('name-bot');
                    name.textContent = bot_name;
                    msg = document.createElement('div');
                    msg.classList.add('msg');
                    box.appendChild(name);
                    box.appendChild(msg);
                    element.appendChild(box);
                } else {
                    msg = box.querySelector('.msg');
                }
                if (msg) {
                    msg.innerHTML+= content;
                }
            }
            highlightCode();
            scrollToBottom();
        }
        function appendNode(content) {
            prevScroll = 0;
            const element = document.getElementById('_nodes_');
            if (element) {
                element.classList.remove('empty_list');
                element.innerHTML += content;
            }
            highlightCode();
            scrollToBottom();
        }
        function appendExtra(id, content) {
            prevScroll = 0;
            const element = document.getElementById('msg-bot-' + id);
            if (element) {
                const extra = element.querySelector('.msg-extra');
                if (extra) {
                    extra.innerHTML+= content;
                }
            }
            highlightCode();
            scrollToBottom();
        }
        function removeNode(id) {
            prevScroll = 0;
            let element = document.getElementById('msg-user-' + id);
            if (element) {
                element.remove();
            }
            element = document.getElementById('msg-bot-' + id);
            if (element) {
                element.remove();
            }
            highlightCode();
            scrollToBottom();
        }
        function removeNodesFromId(id) {
            prevScroll = 0;
            const container = document.getElementById('_nodes_');
            if (container) {
                const elements = container.querySelectorAll('.msg-box');
                remove = false;
                elements.forEach(function(element) {
                    if (element.id.endsWith('-' + id)) {
                        remove = true;
                    }
                    if (remove) {
                        element.remove();
                    }
                });
            }
            highlightCode();
            scrollToBottom();
        }
        function replaceOutput(bot_name, content) {
            const element = document.getElementById('_append_output_');
            if (element) {
                let box = element.querySelector('.msg-box');
                let msg;
                if (!box) {
                    box = document.createElement('div');
                    box.classList.add('msg-box');
                    box.classList.add('msg-bot');
                    const name = document.createElement('div');
                    name.classList.add('name-header');
                    name.classList.add('name-bot');
                    name.textContent = bot_name;
                    msg = document.createElement('div');
                    msg.classList.add('msg');
                    box.appendChild(name);
                    box.appendChild(msg);
                    element.appendChild(box);
                } else {
                    msg = box.querySelector('.msg');
                }
                if (msg) {
                    msg.innerHTML = content;
                }
            }
            highlightCode();
            scrollToBottom();
        }
        function appendToolOutput(content) {
            hideToolOutputLoader();
            enableToolOutput();
            const elements = document.querySelectorAll('.tool-output');
            if (elements.length > 0) {
                const last = elements[elements.length - 1];
                const contentEl = last.querySelector('.content');
                if (contentEl) {
                    contentEl.innerHTML += content;
                }
            }
        }
        function updateToolOutput(content) {
            hideToolOutputLoader();
            enableToolOutput();
            const elements = document.querySelectorAll('.tool-output');
            if (elements.length > 0) {
                const last = elements[elements.length - 1];
                const contentEl = last.querySelector('.content');
                if (contentEl) {
                    contentEl.innerHTML = content;
                }
            }
        }
        function clearToolOutput(content) {
            hideToolOutputLoader();
            enableToolOutput();
            const elements = document.querySelectorAll('.tool-output');
            if (elements.length > 0) {
                const last = elements[elements.length - 1];
                const contentEl = last.querySelector('.content');
                if (contentEl) {
                    contentEl.innerHTML = '';
                }
            }
        }
        function showToolOutputLoader() {
            const elements = document.querySelectorAll('.msg-bot');
            if (elements.length > 0) {
                const last = elements[elements.length - 1];
                const contentEl = last.querySelector('.spinner');
                if (contentEl) {
                    contentEl.style.display = 'inline-block';
                }
            }
        }
        function hideToolOutputLoader() {
            const elements = document.querySelectorAll('.msg-bot');
            // hide all loaders
            if (elements.length > 0) {
                elements.forEach(function(element) {
                    const contentEl = element.querySelector('.spinner');
                    if (contentEl) {
                        contentEl.style.display = 'none';
                    }
                });
            }
        }
        function beginToolOutput() {
            showToolOutputLoader();
        }
        function endToolOutput() {
            hideToolOutputLoader();
        }
        function enableToolOutput() {
            const elements = document.querySelectorAll('.tool-output');
            if (elements.length > 0) {
                const last = elements[elements.length - 1];
                last.style.display = 'block';
            }
        }
        function disableToolOutput() {
            const elements = document.querySelectorAll('.tool-output');
            if (elements.length > 0) {
                const last = elements[elements.length - 1];
                last.style.display = 'none';
            }
        }
        function toggleToolOutput(id) {
            const element = document.getElementById('msg-bot-' + id);
            if (element) {
                const outputEl = element.querySelector('.tool-output');
                if (outputEl) {
                    const contentEl = outputEl.querySelector('.content');
                    if (contentEl.style.display === 'none') {
                        contentEl.style.display = 'block';
                    } else {
                        contentEl.style.display = 'none';
                    }
                    const toggleEl = outputEl.querySelector('.toggle-cmd-output img');
                    if (toggleEl) {
                        toggleEl.classList.toggle('toggle-expanded');
                    }
                }
            }
        }
        function updateFooter(content) {
            const element = document.getElementById('_footer_');
            if (element) {
                element.innerHTML = content;
            }
        }
        function clearNodes() {
            prevScroll = 0;
            const element = document.getElementById('_nodes_');
            if (element) {
                element.textContent = '';
                element.classList.add('empty_list');
            }
        }
        function clearInput() {
            const element = document.getElementById('_append_input_');
            if (element) {
                element.textContent = '';
            }
        }
        function clearOutput() {
            const element = document.getElementById('_append_output_');
            if (element) {
                element.textContent = '';
            }
        }
        function enableEditIcons() {
            const container = document.body;
            if (container) {
                container.classList.add('display-edit-icons');
            }
        }
        function disableEditIcons() {
            const container = document.body;
            if (container) {
                container.classList.remove('display-edit-icons');
            }
        }
        function enableTimestamp() {
            const container = document.body;
            if (container) {
                container.classList.add('display-timestamp');
            }
        }
        function disableTimestamp() {
            const container = document.body;
            if (container) {
                container.classList.remove('display-timestamp');
            }
        }
        function enableBlocks() {
            const container = document.body;
            if (container) {
                container.classList.add('display-blocks');
            }
        }
        function disableBlocks() {
            const container = document.body;
            if (container) {
                container.classList.remove('display-blocks');
            }
        }
        function updateCSS(styles) {
            const style = document.createElement('style');
            style.innerHTML = styles;
            const oldStyle = document.querySelector('style');
            if (oldStyle) {
                oldStyle.remove();
            }
            document.head.appendChild(style);
        }
        function bridgeCopyCode(text) {
            if (bridge) {
                bridge.copy_text(text);
            }
        }
        function bridgeUpdateScrollPosition(pos) {
            if (bridge) {
                bridge.update_scroll_position(pos);
            }
        }
        function getScrollPosition() {
            bridgeUpdateScrollPosition(window.scrollY);
        }
        function setScrollPosition(pos) {
            window.scrollTo(0, pos);
            prevScroll = parseInt(pos);
        }  
        document.addEventListener('DOMContentLoaded', function() {
            const container = document.getElementById('container');
            function addClassToMsg(id, className) {
                const msgElement = document.getElementById('msg-bot-' + id);
                if (msgElement) {
                    msgElement.classList.add(className);
                }
            }
            function removeClassFromMsg(id, className) {
                const msgElement = document.getElementById('msg-bot-' + id);
                if (msgElement) {
                    msgElement.classList.remove(className);
                }
            }
            container.addEventListener('mouseover', function(event) {
                if (event.target.classList.contains('action-img')) {
                    const id = event.target.getAttribute('data-id');
                    addClassToMsg(id, 'msg-highlight');
                }
            });        
            container.addEventListener('mouseout', function(event) {
                if (event.target.classList.contains('action-img')) {
                    const id = event.target.getAttribute('data-id');
                    removeClassFromMsg(id, 'msg-highlight');
                }
            });
            container.addEventListener('click', function(event) {
                if (event.target.classList.contains('code-header-copy')) {
                    event.preventDefault();
                    const parent = event.target.closest('.code-wrapper');
                    const source = parent.querySelector('code');
                    if (source) {
                        const text = source.textContent || source.innerText;
                        bridgeCopyCode(text);
                    }                        
                }
            });
        });
        </script>
        </body>
        </html>
        """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.18 19:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.config import BaseConfig, BasePlugin


class Config(BaseConfig):
    def __init__(self, plugin: BasePlugin = None, *args, **kwargs):
        super(Config, self).__init__(plugin)
        self.plugin = plugin

    def from_defaults(self, plugin: BasePlugin = None):
        """
        Set default options for plugin

        :param plugin: plugin instance
        """
        prompt = "IMAGE ANALYSIS: You are an expert in analyzing photos. Each time I send you a photo, you will " \
                 "analyze it as accurately as you can, answer all my questions, and offer any help on the subject " \
                 "related to the photo.  Remember to always describe in great detail all the aspects related to the " \
                 "photo, try to place them in the context of the conversation, and make a full analysis of what you " \
                 "see. "
        plugin.add_option(
            "model",
            type="combo",
            use="models",
            value="gpt-4o",
            label="Model",
            description="Model used to temporarily providing vision abilities, "
                        "default: gpt-4o",
            tooltip="Model",
        )
        plugin.add_option(
            "prompt",
            type="textarea",
            value=prompt,
            label="Prompt",
            description="Prompt used for vision mode. It will append or replace current system prompt "
                        "when using vision model",
            tooltip="Prompt",
            advanced=True,
        )
        plugin.add_option(
            "replace_prompt",
            type="bool",
            value=False,
            label="Replace prompt",
            description="Replace whole system prompt with vision prompt against appending "
                        "it to the current prompt",
            tooltip="Replace whole system prompt with vision prompt against appending it to the "
                    "current prompt",
            advanced=True,
        )
        plugin.add_cmd(
            "analyze_image_attachment",
            instruction="the function sends the user's image to the vision model and requests an image analysis. Use this function whenever the user asks for an analysis of a given image. The image will be automatically uploaded during the execution of the function.",
            params=[
                {
                    "name": "prompt",
                    "type": "str",
                    "description": "prompt with instructions for image analysis",
                    "required": True,
                },
                {
                    "name": "path",
                    "type": "str",
                    "description": "path to image, if not provided then current image will be used",
                    "required": False,
                },
            ],
            enabled=True,
        )
        plugin.add_cmd(
            "analyze_screenshot",
            instruction="make a screenshot and send it to analyze to vision model",
            params=[
                {
                    "name": "prompt",
                    "type": "str",
                    "description": "prompt with instructions for image analysis",
                    "required": True,
                },
            ],
            enabled=True,
        )
        plugin.add_cmd(
            "analyze_camera_capture",
            instruction="capture image from camera and send it to analyze to vision model",
            params=[
                {
                    "name": "prompt",
                    "type": "str",
                    "description": "prompt with instructions for image analysis",
                    "required": True,
                },
            ],
            enabled=True,
        )
        plugin.add_cmd(
            "camera_capture",
            instruction="capture image from webcam",
            params=[],
            enabled=True,
        )
        plugin.add_cmd(
            "make_screenshot",
            instruction="make desktop screenshot",
            params=[],
            enabled=True,
        )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.23 21:00:00                  #
# ================================================== #

from PySide6.QtCore import Signal, QObject, QRunnable, Slot


class WorkerSignals(QObject):
    success = Signal(str)
    error = Signal(object)


class AttachmentWorker(QRunnable):
    def __init__(self, *args, **kwargs):
        super(AttachmentWorker, self).__init__()
        self.signals = WorkerSignals()
        self.args = args
        self.kwargs = kwargs
        self.window = None
        self.meta = None
        self.mode = None
        self.prompt = ""

    @Slot()
    def run(self):
        """Index attachments"""
        try:
            self.window.controller.chat.attachment.upload(self.meta, self.mode, self.prompt)
            self.signals.success.emit(self.prompt)
        except Exception as e:
            if self.signals is not None:
                self.signals.error.emit(e)
            print("Attachment indexing error", e)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

import json
from typing import List, Dict
from urllib.parse import quote
from .base import BaseProvider


class GoogleCustomSearch(BaseProvider):
    def __init__(self, *args, **kwargs):
        """
        Google Custom Search provider

        :param args: args
        :param kwargs: kwargs
        """
        super(GoogleCustomSearch, self).__init__(*args, **kwargs)
        self.plugin = kwargs.get("plugin")
        self.id = "google_custom_search"
        self.name = "Google"
        self.type = ["search_engine"]

    def init_options(self):
        """Initialize options"""
        url_api = {
            "API Key": "https://developers.google.com/custom-search/v1/overview",
        }
        url_cx = {
            "CX ID": "https://programmablesearchengine.google.com/controlpanel/all",
        }
        self.plugin.add_option(
            "google_api_key",
            type="text",
            value="",
            label="Google Custom Search API KEY",
            description="You can obtain your own API key at "
                        "https://developers.google.com/custom-search/v1/overview",
            tooltip="Google Custom Search API KEY",
            secret=True,
            persist=True,
            tab="google_custom_search",
            urls=url_api,
        )
        self.plugin.add_option(
            "google_api_cx",
            type="text",
            value="",
            label="Google Custom Search CX ID",
            description="You will find your CX ID at https://programmablesearchengine.google.com/controlpanel/all\n"
                        "Remember to enable \"Search on ALL internet pages\" option in project settings.",
            tooltip="Google Custom Search CX ID",
            secret=True,
            persist=True,
            tab="google_custom_search",
            urls=url_cx,
        )

    def search(
            self,
            query: str,
            limit: int = 10,
            offset: int = 0
    ) -> List[str]:
        """
        Execute search query and return list of urls

        :param query: query
        :param limit: limit
        :param offset: offset
        :return: list of urls
        """
        key = self.get_key()
        cx = self.get_cx()

        if limit < 1:
            limit = 1
        if limit > 10:
            limit = 10

        if limit + offset > 100:
            limit = 100 - offset

        urls = []
        url = 'https://www.googleapis.com/customsearch/v1'
        url += '?key=' + str(key)
        url += '&cx=' + str(cx)
        url += '&num=' + str(limit)
        url += '&sort=date-sdate:d:s'
        url += '&fields=items(link)'
        url += '&start=' + str(offset)
        url += '&q=' + quote(query)

        data = self.plugin.get_url(url)
        res = json.loads(data)

        if 'items' not in res:
            return []
        for item in res['items']:
            urls.append(item['link'])

        return urls

    def is_configured(self, cmds: List[Dict]) -> bool:
        """
        Check if provider is configured (required API keys, etc.)

        :param cmds: executed commands list
        :return: True if configured, False if configuration is missing
        """
        required = ["web_search", "web_urls"]
        key = self.get_key()
        cx = self.get_cx()
        need_api_key = False
        for item in cmds:
            if item["cmd"] in required:
                need_api_key = True
                break
        if need_api_key and \
                (key is None or cx is None or key == "" or cx == ""):
            return False
        return True

    def get_config_message(self) -> str:
        """
        Return message to display when provider is not configured

        :return: message
        """
        return "Google Custom Search API key or Google Search CX is not set. Please set credentials in plugin settings."

    def get_key(self) -> str:
        """
        Return Google API key

        :return: Google API key
        """
        return str(self.plugin.get_option_value("google_api_key"))

    def get_cx(self) -> str:
        """
        Return Google API CX

        :return: Google API CX
        """
        return str(self.plugin.get_option_value("google_api_cx"))





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

import os
from typing import List

from PySide6.QtCore import Slot, QObject

from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans


class Image(QObject):
    def __init__(self, window=None):
        """
        Image generation core

        :param window: Window instance
        """
        super().__init__()
        self.window = window

    def install(self):
        """Install provider data, img dir, etc."""
        img_dir = os.path.join(self.window.core.config.get_user_dir("img"))
        if not os.path.exists(img_dir):
            os.makedirs(img_dir, exist_ok=True)

    @Slot(object, list, str)
    def handle_finished(
            self,
            ctx: CtxItem,
            paths: List[str],
            prompt: str
    ):
        """
        Handle finished image generation

        :param ctx: CtxItem
        :param paths: images paths list
        :param prompt: prompt used for generate images
        """
        self.window.controller.chat.image.handle_response(ctx, paths, prompt)

    @Slot(object, list, str)
    def handle_finished_inline(
            self,
            ctx: CtxItem,
            paths: List[str],
            prompt: str
    ):
        """
        Handle finished image generation

        :param ctx: CtxItem
        :param paths: images paths list
        :param prompt: prompt used for generate images
        """
        self.window.controller.chat.image.handle_response_inline(
            ctx,
            paths,
            prompt,
        )

    @Slot()
    def handle_status(self, msg: str):
        """
        Handle thread status message

        :param msg: status message
        """
        self.window.update_status(msg)

        is_log = False
        if self.window.core.config.has("log.dalle") \
                and self.window.core.config.get("log.dalle"):
            is_log = True
        self.window.core.debug.info(msg, not is_log)
        if is_log:
            print(msg)

    @Slot()
    def handle_error(self, msg: any):
        """
        Handle thread error message

        :param msg: error message
        """
        self.window.update_status(msg)
        self.window.core.debug.log(msg)

    def save_image(self, path: str, image: bytes) -> bool:
        """
        Save image to file

        :param path: path to save
        :param image: image data
        :return: True if success
        """
        try:
            with open(path, 'wb') as file:
                file.write(image)
            return True
        except Exception as e:
            print(trans('img.status.save.error') + ": " + str(e))
            return False

    def make_safe_filename(self, name: str) -> str:
        """
        Make safe filename

        :param name: filename to make safe
        :return: safe filename
        """
        def safe_char(c):
            if c.isalnum():
                return c
            else:
                return "_"
        return "".join(safe_char(c) for c in name).rstrip("_")[:30]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

import json
from typing import List, Dict
from urllib.parse import quote

import requests

from .base import BaseProvider


class MicrosoftBingSearch(BaseProvider):
    def __init__(self, *args, **kwargs):
        """
        Google Custom Search provider

        :param args: args
        :param kwargs: kwargs
        """
        super(MicrosoftBingSearch, self).__init__(*args, **kwargs)
        self.plugin = kwargs.get("plugin")
        self.id = "microsoft_bing"
        self.name = "Bing"
        self.type = ["search_engine"]

    def init_options(self):
        """Initialize options"""
        url_api = {
            "API Key": "https://www.microsoft.com/en-us/bing/apis/bing-web-search-api",
        }
        self.plugin.add_option(
            "bing_api_key",
            type="text",
            value="",
            label="Bing Search API KEY",
            description="You can obtain your own API key at "
                        "https://www.microsoft.com/en-us/bing/apis/bing-web-search-api",
            tooltip="Bing Search API KEY",
            secret=True,
            persist=True,
            tab="microsoft_bing",
            urls=url_api,
        )
        self.plugin.add_option(
            "bing_endpoint",
            type="text",
            value="https://api.bing.microsoft.com/v7.0/search",
            label="Bing Search API endpoint",
            description="API endpoint for Bing Search API",
            tooltip="Bing Search API endpoint",
            persist=False,
            tab="microsoft_bing",
        )

    def search(
            self,
            query: str,
            limit: int = 10,
            offset: int = 0
    ) -> List[str]:
        """
        Execute search query and return list of urls

        :param query: query
        :param limit: limit
        :param offset: offset
        :return: list of urls
        """
        key = self.get_key()
        urls = []

        if limit < 1:
            limit = 1
        if limit > 10:
            limit = 10

        if limit + offset > 100:
            limit = 100 - offset

        headers = {
            'Ocp-Apim-Subscription-Key': key,
        }
        endpoint = self.plugin.get_option_value("bing_endpoint")
        url = f"{endpoint}?q={quote(query)}&count={limit}&offset={offset}&mkt=en-US"
        response = requests.get(
            url,
            headers=headers,
        )
        if response.status_code == 200:
            data = response.json()
            for result in data['webPages']['value']:
                urls.append(result['url'])
        else:
            print('Error:', response.status_code, response.text)
        return urls

    def is_configured(self, cmds: List[Dict]) -> bool:
        """
        Check if provider is configured (required API keys, etc.)

        :param cmds: executed commands list
        :return: True if configured, False if configuration is missing
        """
        required = ["web_search", "web_urls"]
        key = self.get_key()
        need_api_key = False
        for item in cmds:
            if item["cmd"] in required:
                need_api_key = True
                break
        if need_api_key and (key is None or key == ""):
            return False
        return True

    def get_config_message(self) -> str:
        """
        Return message to display when provider is not configured

        :return: message
        """
        return "Microsoft Bing Search API key is not set. Please set credentials in plugin settings."

    def get_key(self) -> str:
        """
        Return Google API key

        :return: Google API key
        """
        return str(self.plugin.get_option_value("bing_api_key"))





# !/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.02.01 16:00:00                  #
# ================================================== #

from typing import Any, Optional

from pygpt_net.core.events import AppEvent, RenderEvent
from pygpt_net.core.tabs.tab import Tab
from pygpt_net.item.ctx import CtxMeta
from pygpt_net.utils import trans


class Tabs:
    def __init__(self, window=None):
        """
        UI tabs controller

        :param window: Window instance
        """
        self.window = window
        self.active_idx = 0
        self.prev_idx = 0
        self.initialized = False
        self.appended = False
        self.current = 0
        self.column_idx = 0
        self.tmp_column_idx = 0
        self.locked = False
        self.col = {}

    def setup(self):
        """Setup tabs"""
        self.window.core.tabs.load()
        self.window.controller.notepad.load()
        self.setup_options()
        self.initialized = True

    def setup_options(self):
        """Setup options"""
        state = self.window.core.config.get("layout.split", False)
        self.window.ui.nodes['layout.split'].setChecked(state)
        if not state:
            self.window.ui.splitters['columns'].setSizes([1, 0])

    def add(
            self,
            type: int,
            title: str,
            icon: Optional[str] = None,
            child: Any = None,
            data_id: Optional[int] = None,
            tool_id: Optional[str] = None,
    ):
        """
        Add a new tab

        :param type: Tab type
        :param title: Tab title
        :param icon: Tab icon
        :param child: Tab child (child widget)
        :param data_id: Tab data ID (child data ID)
        :param tool_id: Tool ID
        """
        self.window.core.tabs.add(
            type=type,
            title=title,
            icon=icon,
            child=child,
            data_id=data_id,
            tool_id=tool_id
        )

    def append(
            self,
            type: int,
            tool_id: Optional[str] = None,
            idx: int = 0,
            column_idx: int = 0
    ):
        """
        Append tab at tab index

        :param type: Tab type
        :param tool_id: Tool ID
        :param idx: Tab index
        :param column_idx: Column index
        """
        self.appended = True  # lock reload in previous tab
        self.column_idx = column_idx  # switch to column
        tab = self.window.core.tabs.append(
            type=type,
            idx=idx,
            column_idx=column_idx,
            tool_id=tool_id
        )
        self.switch_tab_by_idx(tab.idx, column_idx)  # switch to new tab

    def reload_titles(self):
        """Reload tab titles"""
        self.window.core.tabs.reload_titles()

    def update_current(self):
        """Update current tab"""
        curr_tab = self.get_current_tab()
        curr_column = self.get_current_column_idx()
        if curr_column not in self.col:
            self.col[curr_column] = -1
        if curr_tab is not None:
            self.col[curr_column] = curr_tab.pid

    def reload(self):
        """Reload tabs"""
        self.window.core.tabs.reload()
        event = RenderEvent(RenderEvent.PREPARE)
        self.window.dispatch(event)

    def reload_after(self):
        """Reload tabs after"""
        for pid in self.window.ui.nodes['output']:
            if self.window.core.config.get("render.plain") is True:
                self.window.ui.nodes['output_plain'][pid].setVisible(True)
                self.window.ui.nodes['output'][pid].setVisible(False)
            else:
                self.window.ui.nodes['output_plain'][pid].setVisible(False)
                self.window.ui.nodes['output'][pid].setVisible(True)
        #self.switch_tab(Tab.TAB_CHAT)

    def on_tab_changed(
            self,
            idx: int,
            column_idx: int = 0
    ):
        """
        Output tab changed

        :param idx: tab index
        :param column_idx: column index
        """
        tab = self.window.core.tabs.get_tab_by_index(idx, column_idx)
        if tab is None:
            self.appended = False
            return

        if self.appended:
            self.appended = False
            if tab.type == Tab.TAB_CHAT:
                self.current = idx
                meta = self.window.controller.ctx.new()  # new context
                if meta is not None:
                    self.window.controller.ctx.load(meta.id)  # reload

        prev_tab = self.current
        prev_column = self.column_idx
        self.current = idx
        self.column_idx = column_idx
        self.window.controller.ui.mode.update()
        self.window.controller.ui.vision.update()

        # check type
        if tab.type == Tab.TAB_NOTEPAD:
            self.window.controller.notepad.opened_once = True
            self.window.controller.notepad.on_open(idx, column_idx)
        elif tab.type == Tab.TAB_CHAT:
            # get meta for selected tab, if not loaded yet then append meta here
            meta_id = self.window.core.ctx.output.prepare_meta(tab)
            meta = self.window.core.ctx.get_meta_by_id(meta_id)
            if meta is not None:
                self.window.controller.ctx.load(meta.id)  # reload renderer
        elif tab.type == Tab.TAB_TOOL_PAINTER:
            if self.window.core.config.get('vision.capture.enabled'):
                self.window.controller.camera.enable_capture()
        elif tab.type == Tab.TAB_TOOL_CALENDAR:
            self.window.controller.calendar.update()
            self.window.controller.calendar.update_ctx_counters()

        if prev_tab != idx or prev_column != column_idx:
            self.window.dispatch(AppEvent(AppEvent.TAB_SELECTED))  # app event

        self.on_changed()

        self.window.controller.ui.update()
        self.update_current()

    def on_changed(self):
        """On Tab or column changed event (any)"""
        tab = self.get_current_tab()
        if tab is None:
            return
        self.window.controller.audio.on_tab_changed(tab)

    def get_current_idx(self, column_idx: int = 0) -> int:
        """
        Get current tab index

        :param column_idx: column index
        :return: tab index
        """
        return self.current

    def get_current_column_idx(self) -> int:
        """
        Get current column index

        :return: column index
        """
        return self.column_idx

    def get_current_tab(self) -> Optional[Tab]:
        """
        Get current tab

        :return: tab
        """
        return self.window.core.tabs.get_tab_by_index(self.get_current_idx(), self.column_idx)

    def get_current_type(self) -> Optional[int]:
        """
        Get current tab type

        :return: tab type
        """
        tab = self.window.core.tabs.get_tab_by_index(self.get_current_idx(), self.column_idx)

        if tab is None:
            return None
        return tab.type

    def get_current_pid(self) -> Optional[int]:
        """
        Get current tab PID

        :return: tab PID
        """
        tab = self.window.core.tabs.get_tab_by_index(self.get_current_idx(), self.column_idx)
        if tab is None:
            return None
        return tab.pid

    def get_type_by_idx(self, idx: int) -> Optional[int]:
        """
        Get tab type by index

        :param idx: tab index
        :return: tab type
        """
        tab = self.window.core.tabs.get_tab_by_index(idx, self.column_idx)
        if tab is None:
            return None
        return tab.type

    def get_first_idx_by_type(self, type: int) -> Optional[int]:
        """
        Get first tab index by type

        :param type: tab type
        :return: tab index
        """
        return self.window.core.tabs.get_min_idx_by_type(type, self.column_idx)

    def on_column_changed(self):
        """Column changed event"""
        if self.locked:
            return
        tabs = self.window.ui.layout.get_tabs_by_idx(self.column_idx)
        tabs.set_active(True)

        if self.column_idx == 0:
            second_tabs = self.window.ui.layout.get_tabs_by_idx(1)
        else:
            second_tabs = self.window.ui.layout.get_tabs_by_idx(0)
        second_tabs.set_active(False)
        idx = tabs.currentIndex()
        self.current = idx
        tab = self.window.core.tabs.get_tab_by_index(self.current, self.column_idx)
        if tab is None:
            return
        current_ctx = self.window.core.ctx.get_current()
        if (current_ctx is not None and current_ctx != tab.data_id) or current_ctx is None:
            self.window.controller.ctx.select_on_list_only(tab.data_id)
        self.window.controller.ui.update()
        self.update_current()

    def on_tab_clicked(
            self,
            idx: int,
            column_idx: int = 0
    ):
        """
        Tab click event

        :param idx: tab index
        :param column_idx: column index
        """
        self.current = idx
        self.column_idx = column_idx
        self.on_column_changed()
        self.on_changed()
        self.update_current()

    def on_column_focus(self, idx: int):
        """
        Column focus event

        :param idx: column index
        """
        if self.column_idx == idx:
            return
        self.column_idx = idx
        self.on_column_changed()
        self.on_changed()
        self.update_current()

    def on_tab_dbl_clicked(
            self,
            idx: int,
            column_idx: int = 0
    ):
        """
        Tab double click event

        :param idx: tab index
        :param column_idx: column index
        """
        self.column_idx = column_idx
        self.on_tab_changed(idx, column_idx)
        self.update_current()

    def on_tab_closed(
            self,
            idx: int,
            column_idx: int = 0
    ):
        """
        Tab close event

        :param idx: tab index
        :param column_idx: column index
        """
        if self.locked:
            return
        self.window.core.tabs.remove_tab_by_idx(idx, column_idx)
        self.on_changed()
        self.update_current()

    def on_tab_moved(
            self,
            idx: int,
            column_idx: int = 0
    ):
        """
        Tab moved event

        :param idx: tab index
        :param column_idx: column index
        """
        if self.locked:
            return
        self.window.core.tabs.update()
        self.update_current()

    def close(
            self,
            idx: int,
            column_idx: int = 0
    ):
        """
        Close tab

        :param idx: tab index
        :param column_idx: column index
        """
        self.on_tab_closed(idx, column_idx)
        self.update_current()

    def close_all(
            self,
            type: int,
            column_idx: int = 0,
            force: bool = False
    ):
        """
        Close all tabs

        :param type: tab type
        :param column_idx: column index
        :param force: force close
        """
        if not force:
            self.tmp_column_idx = column_idx
            self.window.ui.dialogs.confirm(
                type='tab.close_all',
                id=type,
                msg=trans('tab.close_all.confirm'),
            )
            return
        column_idx = self.tmp_column_idx
        self.window.core.tabs.remove_all_by_type(type, column_idx)
        self.on_changed()
        self.update_current()

    def next_tab(self):
        """Switch to next tab"""
        tabs = self.window.ui.layout.get_active_tabs()
        current = tabs.currentIndex()
        all = len(tabs.children())
        next = current + 1
        if next >= all:
            next = 0
        self.switch_tab_by_idx(next)

    def prev_tab(self):
        """Switch to previous tab"""
        tabs = self.window.ui.layout.get_active_tabs()
        current = tabs.currentIndex()
        all = len(tabs.children())
        prev = current - 1
        if prev < 0:
            prev = all - 1
        self.switch_tab_by_idx(prev)

    def switch_tab(self, type: int):
        """
        Switch tab

        :param type: tab type
        """
        idx = self.window.core.tabs.get_min_idx_by_type(type)
        if idx is not None:
            self.switch_tab_by_idx(idx)

    def switch_tab_by_idx(
            self,
            idx: int,
            column_idx: int = 0
    ):
        """
        Switch tab by index

        :param idx: tab index
        :param column_idx: column index
        """
        tabs = self.window.ui.layout.get_tabs_by_idx(column_idx)
        tabs.setCurrentIndex(idx)
        self.on_tab_changed(idx, column_idx)

    def get_current_tab_name(self) -> str:
        """
        Get current tab name

        :return: tab name
        """
        tabs = self.window.ui.layout.get_active_tabs()
        return tabs.tabText(self.current)

    def get_current_tab_name_for_audio(self) -> str:
        """
        Get current tab name for audio description

        :return: tab name
        """
        tab = self.get_current_tab()
        if tab is None:
            return ""

        title = ""
        if tab.type in self.window.core.tabs.titles:
            title = trans(self.window.core.tabs.titles[tab.type])

        # if more than 1 with this type then attach position info
        num = self.window.core.tabs.count_by_type(tab.type)
        if num > 1:
            order = self.window.core.tabs.get_order_by_idx_and_type(tab.idx, tab.type)
            if order != -1:
                title += " #" + str(order)
        if tab.tooltip is not None and tab.tooltip != "":
            title += " - " + tab.tooltip
        return title

    def update_tooltip(self, tooltip: str):
        """
        Update tab tooltip

        :param tooltip: tooltip text
        """
        tabs = self.window.ui.layout.get_active_tabs()
        tabs.setTabToolTip(self.current, tooltip)

    def rename(
            self,
            idx: int,
            column_idx: int = 0
    ):
        """
        Rename tab (show dialog)

        :param idx: tab idx
        :param column_idx: column idx
        """
        # get tab
        tab = self.window.core.tabs.get_tab_by_index(idx, column_idx)
        if tab is None:
            return
        # set dialog and show
        self.window.ui.dialog['rename'].id = 'tab'
        self.window.ui.dialog['rename'].input.setText(tab.title)
        self.window.ui.dialog['rename'].current = idx
        self.window.ui.dialog['rename'].show()

    def update_name(
            self,
            idx: int,
            name: str,
            close: bool = True
    ):
        """
        Update tab title

        :param idx: tab idx
        :param name: new title
        :param close: close dialog
        """
        self.window.core.tabs.update_title(idx, name)
        if close:
            self.window.ui.dialog['rename'].close()

    def update_current_name(self, name: str):
        """
        Update current tab title

        :param name: new title
        """
        self.update_name(self.current, name)

    def update_title(
            self,
            idx: int,
            title: str
    ):
        """
        Update tab title

        :param idx: tab idx
        :param title: new title
        """
        # check if current tab is chat
        if self.get_current_type() != Tab.TAB_CHAT:
            return
        tabs = self.window.ui.layout.get_active_tabs()
        tooltip = title
        tabs.setTabToolTip(idx, tooltip)
        if len(title) > 8:
            title = title[:8] + '...'  # truncate to max 8 chars
        self.window.core.tabs.update_title(idx, title, tooltip)

    def update_title_current(self, title: str):
        """
        Update current tab title

        :param title: new title
        """
        self.update_title(self.current, title)

    def on_load_ctx(self, meta: CtxMeta):
        """
        Load context

        :param meta: context meta
        """
        # get current tab
        tab = self.get_current_tab()
        if tab is not None and tab.type == Tab.TAB_CHAT:
            tab.data_id = meta.id
        self.update_title_current(meta.name)

    def open_by_type(self, type: int):
        """
        Open first tab by type

        :param type: tab type
        """
        idx = self.window.core.tabs.get_min_idx_by_type(type)
        if idx is not None:
            self.switch_tab_by_idx(idx)

    def new_tab(self, column_idx: int = 0):
        """
        Handle [+] button

        :param column_idx: column index
        """
        idx = self.get_current_idx(column_idx)
        self.append(
            type=Tab.TAB_CHAT,
            tool_id=None,
            idx=idx,
            column_idx=column_idx
        )

    def restore_data(self):
        """Restore tab data"""
        data = self.window.core.config.get("tabs.opened", [])
        if not data:
            self.switch_tab_by_idx(0, 0)
            return

        # reverse order, second column is first
        data = dict(reversed(list(data.items())))
        for col_idx in data:
            tab_idx = data[col_idx]
            self.switch_tab_by_idx(int(tab_idx), int(col_idx))

        # set default column to 0
        self.column_idx = 0
        self.on_column_changed()

    def move_tab(
            self,
            idx: int,
            column_idx: int,
            new_column_idx: int
    ):
        """
        Move tab to another column

        :param idx: tab index
        :param column_idx: column index
        :param new_column_idx: new column index
        """
        self.locked = True
        tab = self.window.core.tabs.get_tab_by_index(idx, column_idx)
        self.window.core.tabs.move_tab(tab, new_column_idx)
        self.locked = False
        # switch to new column
        self.column_idx = new_column_idx
        self.on_column_changed()
        # switch to new tab
        self.switch_tab_by_idx(tab.idx, new_column_idx)

    def toggle_split_screen(self, state):
        """
        Toggle split screen mode

        :param state: True if split screen is enabled
        """
        if state:
            # self.rightWidget.show()
            self.window.ui.splitters['columns'].setSizes([1, 1])
        else:
            # self.rightWidget.hide()
            self.window.ui.splitters['columns'].setSizes([1, 0])
            # set to first column
            self.column_idx = 0
            self.on_column_changed()
        self.window.core.config.set("layout.split", state)
        self.window.core.config.save()

    def is_current_by_type(self, type: int) -> bool:
        """
        Check if one of current tabs is of given type

        :param type: tab type
        :return: True if one of tab is of given type
        """
        for col in self.col:
            pid = self.col[col]
            tab = self.window.core.tabs.get_tab_by_pid(pid)
            if tab is not None and tab.type == type:
                return True

    def is_current_tool(self, tool_id: str) -> bool:
        """
        Check if one of current tabs is of given tool ID

        :param tool_id: tool ID
        :return: True if one of tab is of given tool ID
        """
        for col in self.col:
            pid = self.col[col]
            tab = self.window.core.tabs.get_tab_by_pid(pid)
            if tab is not None and tab.tool_id == tool_id:
                return True

    def switch_to_first_chat(self):
        """Switch to first chat tab"""
        if self.is_current_by_type(Tab.TAB_CHAT):
            return
        # abort if active tab is chat
        if self.get_current_type() == Tab.TAB_CHAT:
            return
        # find first chat tab
        for col in self.col:
            pid = self.col[col]
            tab = self.window.core.tabs.get_tab_by_pid(pid)
            if tab is not None and tab.type == Tab.TAB_CHAT:
                self.switch_tab_by_idx(tab.idx, col)
                return



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

import uuid
from typing import Dict

from packaging.version import Version

from pygpt_net.item.assistant import AssistantFileItem
from pygpt_net.provider.core.assistant_file.base import BaseProvider

from .patch import Patch
from .storage import Storage


class DbSqliteProvider(BaseProvider):
    def __init__(self, window=None):
        super(DbSqliteProvider, self).__init__(window)
        self.window = window
        self.patcher = Patch(window, self)
        self.storage = Storage(window)
        self.id = "db_sqlite"
        self.type = "assistant_file"

    def attach(self, window):
        """
        Attach window instance

        :param window: Window instance
        """
        self.window = window
        self.storage.attach(window)

    def patch(self, version: Version) -> bool:
        """
        Patch versions

        :param version: current app version
        :return: True if migrated
        """
        return self.patcher.execute(version)

    def create_id(self) -> str:
        """
        Create unique uuid

        :return: uuid
        """
        return str(uuid.uuid4())

    def create(self, file: AssistantFileItem) -> int:
        """
        Create new and return its ID

        :param file: AssistantFileItem
        :return: file ID
        """
        if file.record_id is None or file.record_id == "":
            file.uuid = self.create_id()
            file.record_id = self.storage.insert(file)
        return file.record_id

    def load_all(self) -> Dict[str, AssistantFileItem]:
        """
        Load files from DB

        :return: files dict
        """
        return self.storage.get_all()

    def load(self, id: int) -> AssistantFileItem:
        """
        Load file from DB

        :param id: file ID
        :return: file item
        """
        return self.storage.get_by_id(id)

    def save(self, file: AssistantFileItem):
        """
        Save file to DB

        :param file: AssistantFileItem
        """
        try:
            self.storage.save(file)
        except Exception as e:
            self.window.core.debug.log(e)
            print("Error while saving filed: {}".format(str(e)))

    def save_all(self, items: Dict[str, AssistantFileItem]):
        """
        Save all files to DB

        :param items: dict of AssistantFileItem objects
        """
        try:
            for id in items:
                file = items[id]
                self.storage.save(file)
        except Exception as e:
            self.window.core.debug.log(e)
            print("Error while saving file: {}".format(str(e)))

    def get_by_store_or_thread(self, store_id: str, thread_id: str) -> dict:
        """
        Return dict with AssistantFileItem objects, indexed by record ID

        :return: dict of AssistantFileItem objects
        """
        return self.storage.get_by_store_or_thread(store_id, thread_id)

    def count_by_store_or_thread(self, store_id: str, thread_id: str) -> int:
        """
        Count AssistantFileItem objects, indexed by ID

        :return: number of files
        """
        return self.storage.count_by_store_or_thread(store_id, thread_id)

    def get_all_by_file_id(self, file_id: str) -> dict:
        """
        Get all files by file ID

        :param file_id: file ID
        :return: files dict
        """
        return self.storage.get_all_by_file_id(file_id)

    def delete_by_id(self, id: int) -> bool:
        """
        Delete file by ID

        :param id: file ID
        :return: True if deleted
        """
        return self.storage.delete_by_id(id)

    def delete_by_file_id(self, file_id: str) -> bool:
        """
        Delete file by file ID

        :param file_id: file ID
        :return: True if deleted
        """
        return self.storage.delete_by_file_id(file_id)

    def clear_store_from_files(self, store_id: str) -> bool:
        """
        Clear store from files

        :param store_id: store ID
        :return: True if deleted
        """
        return self.storage.clear_store_from_files(store_id)

    def clear_all_stores_from_files(self) -> bool:
        """
        Clear all stores from files

        :return: True if deleted
        """
        return self.storage.clear_all_stores_from_files()

    def rename_file(self, record_id: int, name: str) -> bool:
        """
        Rename file

        :param record_id: record ID
        :param name: new name
        :return: True if renamed
        """
        return self.storage.rename_file(record_id, name)

    def truncate_all(self) -> bool:
        """
        Truncate all files

        :return: True if truncated
        """
        return self.storage.truncate_all()

    def truncate_by_store(self, store_id: str) -> bool:
        """
        Truncate all files

        :param store_id: store ID
        :return: True if truncated
        """
        return self.storage.truncate_by_store(store_id)





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.18 03:00:00                  #
# ================================================== #

import os.path
import time
import speech_recognition as sr
import audioop

from PySide6.QtCore import Slot, Signal

from pygpt_net.core.tabs.tab import Tab
from pygpt_net.utils import trans
from pygpt_net.plugin.base.worker import BaseWorker, BaseSignals


class WorkerSignals(BaseSignals):
    transcribed = Signal(str, str)
    on_realtime = Signal(str)


class Worker(BaseWorker):
    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = WorkerSignals()
        self.args = args
        self.kwargs = kwargs
        self.plugin = None
        self.path = None
        self.advanced = False
        self.transcribe = False

    @Slot()
    def run(self):
        """Run worker."""

        # from file
        if self.transcribe:
            self.handle_file()

        # from microphone
        else:
            if self.advanced:
                self.handle_advanced()
            else:
                self.handle_simple()

    def handle_file(self):
        """Handle file"""
        try:
            # do transcribe
            if os.path.exists(self.path):
                transcript = self.plugin.get_provider().transcribe(self.path)
                if transcript is not None and transcript.strip() != '':
                    self.signals.transcribed.emit(self.path, transcript)
                else:
                    self.status('Error: No transcript.')
        except Exception as e:
            self.plugin.window.ui.nodes['audio.transcribe.status'].setText("Failed. Error: {}".format(e))
            self.error(e)
            self.stopped()
            self.status('Error: {}'.format(e))

    def handle_simple(self):
        """Handle mic simple mode."""
        try:
            # do transcribe
            if os.path.exists(self.path):
                # set status
                self.status(trans('audio.speak.wait'))

                # if multimodal audio, then only return path to audio file and do not transcribe
                tab = self.window.controller.ui.tabs.get_current_tab()
                if tab.type == Tab.TAB_CHAT:
                    if self.plugin.window.controller.chat.audio.enabled():
                        self.signals.on_realtime.emit(self.path)
                        self.status('')
                        return

                # transcribe audio
                transcript = self.plugin.get_provider().transcribe(self.path)
                self.status('')

                # handle transcript
                if transcript is not None and transcript.strip() != '':
                    self.response(transcript)

        except Exception as e:
            self.error(e)
            self.stopped()
            self.status('Error: {}'.format(e))

    def handle_advanced(self):
        """Handle mic advanced mode."""
        try:
            if not self.plugin.listening:
                return

            # print("Starting audio listener....")

            self.started()
            self.status('')

            with sr.Microphone() as source:
                while self.plugin.listening and not self.plugin.window.is_closing:
                    self.status('')

                    if self.plugin.stop:
                        self.stopped()
                        self.status('Stop.')
                        break

                    if not self.plugin.can_listen():
                        time.sleep(0.5)
                        continue

                    try:
                        recognizer = sr.Recognizer()

                        # set recognizer options
                        recognizer.energy_threshold = self.plugin.get_option_value(
                            'recognition_energy_threshold'
                        )
                        recognizer.dynamic_energy_threshold = \
                            self.plugin.get_option_value(
                                'recognition_dynamic_energy_threshold'
                            )
                        recognizer.dynamic_energy_adjustment_damping = \
                            self.plugin.get_option_value(
                                'recognition_dynamic_energy_adjustment_damping'
                            )
                        recognizer.dynamic_energy_adjustment_ratio = \
                            self.plugin.get_option_value(
                                'recognition_dynamic_energy_adjustment_ratio'
                            )
                        recognizer.pause_threshold = self.plugin.get_option_value(
                            'recognition_pause_threshold'
                        )
                        adjust_duration = self.plugin.get_option_value(
                            'recognition_adjust_for_ambient_noise_duration'
                        )

                        # adjust for ambient noise
                        if self.plugin.get_option_value('adjust_noise'):
                            recognizer.adjust_for_ambient_noise(
                                source,
                                duration=adjust_duration,
                            )
                            self.plugin.is_first_adjust = False

                        timeout = self.plugin.get_option_value('timeout')
                        phrase_length = self.plugin.get_option_value('phrase_length')

                        # check for magic word, if no magic word detected,
                        # then set to magic word timeout and length
                        if self.plugin.get_option_value('magic_word'):
                            if not self.plugin.magic_word_detected:
                                timeout = self.plugin.get_option_value(
                                    'magic_word_timeout'
                                )
                                phrase_length = self.plugin.get_option_value(
                                    'magic_word_phrase_length'
                                )

                        # set begin status
                        if self.plugin.can_listen():
                            if self.plugin.get_option_value('magic_word'):
                                if self.plugin.magic_word_detected:
                                    self.status(trans('audio.speak.now'))
                                else:
                                    self.status(trans('audio.magic_word.please'))
                            else:
                                self.status(trans('audio.speak.now'))

                        min_energy = self.plugin.get_option_value('min_energy')
                        ambient_noise_energy = min_energy * recognizer.energy_threshold

                        if timeout > 0 and phrase_length > 0:
                            audio_data = recognizer.listen(
                                source,
                                timeout,
                                phrase_length,
                            )
                        elif timeout > 0:
                            audio_data = recognizer.listen(
                                source,
                                timeout,
                            )
                        else:
                            audio_data = recognizer.listen(source)

                        if not self.plugin.can_listen():
                            continue

                        # transcript audio
                        raw_data = audio_data.get_wav_data()
                        is_stop_word = False

                        if raw_data:
                            # check RMS / energy
                            rms = audioop.rms(raw_data, 2)
                            if min_energy > 0:
                                self.status("{}: {} / {} (x{})".format(
                                    trans('audio.speak.energy'),
                                    rms,
                                    int(ambient_noise_energy),
                                    min_energy,
                                ))
                            if rms < ambient_noise_energy:
                                continue

                            # save audio file
                            with open(self.path, "wb") as audio_file:
                                audio_file.write(raw_data)

                            # do transcribe
                            if os.path.exists(self.path):

                                # set status
                                self.status(trans('audio.speak.wait'))

                                # transcribe audio
                                transcript = self.plugin.get_provider().transcribe(self.path)

                                # handle transcript
                                if transcript is not None and transcript.strip() != '':
                                    # fix if empty phrase
                                    is_empty_phrase = False
                                    transcript_check = transcript.strip().lower()
                                    for phrase in self.plugin.empty_phrases:
                                        phrase_check = phrase.strip().lower()
                                        if phrase_check in transcript_check:
                                            is_empty_phrase = True
                                            break

                                    if is_empty_phrase:
                                        continue

                                    if self.plugin.can_listen():
                                        self.response(transcript)

                                    # stop listening if not continuous mode or stop word detected
                                    stop_words = self.plugin.get_words('stop_words')
                                    if len(stop_words) > 0:
                                        is_stop_word = transcript.replace('.', '').strip().lower() in stop_words

                        if not self.plugin.get_option_value('continuous_listen') \
                                or is_stop_word:
                            self.stopped()
                            self.status('')  # clear status
                            break

                    except Exception as e:
                        print("Speech recognition error: {}".format(str(e)))

            self.destroyed()

        except Exception as e:
            self.error(e)
            self.destroyed()
            print("Audio input thread error: {}".format(str(e)))


class ControlWorker(BaseWorker):
    def __init__(self, *args, **kwargs):
        super(ControlWorker, self).__init__()
        self.signals = WorkerSignals()
        self.window = None
        self.args = args
        self.kwargs = kwargs
        self.path = None
        self.transcribe = False

    @Slot()
    def run(self):
        """Handle mic simple mode."""
        try:
            # do transcribe
            if os.path.exists(self.path):
                # set status
                self.status(trans('audio.speak.wait'))
                # transcribe audio
                transcript = self.window.core.plugins.get('audio_input').get_provider().transcribe(self.path)
                self.status('')

                # handle transcript
                if transcript is not None and transcript.strip() != '':
                    self.response(transcript)

        except Exception as e:
            self.error(e)
            self.stopped()
            self.status('Error: {}'.format(e))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.16 01:00:00                  #
# ================================================== #

import os
from datetime import datetime
from typing import Optional
from urllib.parse import urlparse

from PySide6.QtGui import QImage
from PySide6.QtWidgets import QFileDialog, QApplication

from pygpt_net.core.types import (
    MODE_VISION,
)
from pygpt_net.core.events import AppEvent, KernelEvent
from pygpt_net.item.attachment import AttachmentItem
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans


class Attachment:
    def __init__(self, window=None):
        """
        Attachment controller

        :param window: Window instance
        """
        self.window = window
        self.is_lock = False
        self.is_consumed = False

    def setup(self):
        """Setup attachments"""
        # send clear
        if self.window.core.config.has('attachments_send_clear') \
                and self.window.core.config.get('attachments_send_clear'):
            self.window.ui.nodes['attachments.send_clear'].setChecked(True)
        else:
            self.window.ui.nodes['attachments.send_clear'].setChecked(False)

        #  capture clear
        if self.window.core.config.has('attachments_capture_clear') \
                and self.window.core.config.get('attachments_capture_clear'):
            self.window.ui.nodes['attachments.capture_clear'].setChecked(True)
        else:
            self.window.ui.nodes['attachments.capture_clear'].setChecked(False)

        #  auto-index
        if self.window.core.config.has('attachments_auto_index') \
                and self.window.core.config.get('attachments_auto_index'):
            self.window.ui.nodes['attachments.auto_index'].setChecked(True)
        else:
            self.window.ui.nodes['attachments.auto_index'].setChecked(False)

        self.window.core.attachments.load()
        self.update()

    def update(self):
        """Update attachments list"""
        mode = self.window.core.config.get('mode')
        items = self.window.core.attachments.get_all(mode)
        self.window.ui.chat.input.attachments.update(items)
        self.update_tab(mode)

        if not self.has(mode):
            self.window.controller.chat.vision.unavailable()
        else:
            if mode == MODE_VISION or self.window.controller.plugins.is_type_enabled('vision'):
                self.window.controller.chat.vision.available()

        # update tokens counter (vision plugin, etc.)
        self.window.controller.ui.update_tokens()

    def update_tab(self, mode: str):
        """
        Update tab label

        :param mode: mode
        """
        num_files = self.window.core.attachments.count(mode)
        suffix = ''
        if num_files > 0:
            suffix = f' ({num_files})'
        self.window.ui.tabs['input'].setTabText(
            1,
            trans('attachments.tab') + suffix,
        )

    def select(self, mode: str, idx: int):
        """
        Select attachment

        :param mode: mode
        :param idx: index
        """
        self.window.core.attachments.current = self.window.core.attachments.get_id_by_idx(
            mode=mode,
            idx=idx,
        )

    def selection_change(self):
        """
        Select on list change
        """
        # TODO: implement this
        pass

    def delete(
            self,
            idx: int,
            force: bool = False,
            remove_local: bool = False
    ):
        """
        Delete attachment

        :param idx: index of attachment
        :param force: force delete
        :param remove_local: remove local file
        """
        mode = self.window.core.config.get('mode')
        if not force:
            self.window.ui.dialogs.confirm(
                type='attachments.delete',
                id=idx,
                msg=trans('attachments.delete.confirm'),
            )
            return

        file_id = self.window.core.attachments.get_id_by_idx(
            mode=mode,
            idx=idx,
        )
        self.window.core.attachments.delete(
            mode=mode,
            id=file_id,
            remove_local=remove_local,
        )

        # clear current if current == deleted
        if self.window.core.attachments.current == file_id:
            self.window.core.attachments.current = None

        if not self.has(mode):
            self.window.controller.chat.vision.unavailable()

        self.update()

    def rename(self, mode: str, idx: int):
        """
        Rename attachment

        :param mode: mode
        :param idx: selected attachment index
        """

        # get attachment ID by index
        file_id = self.window.core.attachments.get_id_by_idx(
            mode=mode,
            idx=idx,
        )

        # get attachment object by ID
        data = self.window.core.attachments.get_by_id(
            mode=mode,
            id=file_id,
        )
        if data is None:
            return

        # set dialog and show
        self.window.ui.dialog['rename'].id = 'attachment'
        self.window.ui.dialog['rename'].input.setText(data.name)
        self.window.ui.dialog['rename'].current = file_id
        self.window.ui.dialog['rename'].show()
        self.update()

    def update_name(self, file_id: str, name: str):
        """
        Update name

        :param file_id: file_id
        :param name: name
        """
        # rename filename in attachments
        mode = self.window.core.config.get('mode')
        self.window.core.attachments.rename_file(
            mode=mode,
            id=file_id,
            name=name,
        )

        # close rename dialog and update attachments list
        self.window.ui.dialog['rename'].close()
        self.update()

    def add(
            self,
            mode: str,
            attachment: AttachmentItem
    ):
        """
        Add attachment item to list

        :param mode: mode
        :param attachment: attachment object
        """
        self.window.core.attachments.add(
            mode=mode,
            item=attachment,
        )
        self.is_consumed = False  # reset consumed flag
        self.update()

    def clear(
            self,
            force: bool = False,
            remove_local: bool = False,
            auto: bool = False
    ):
        """
        Clear attachments list

        :param force: force clear
        :param remove_local: remove local copies
        :param auto: auto clear
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='attachments.clear',
                id=-1,
                msg=trans('attachments.clear.confirm'),
            )
            return

        # delete all from attachments for current mode
        mode = self.window.core.config.get('mode')
        self.window.core.attachments.delete_all(
            mode=mode,
            remove_local=remove_local,
            auto=auto,
            force=force,
        )
        self.window.controller.chat.vision.unavailable()  # set no content to provide
        self.update()
        if not auto:
            self.window.dispatch(AppEvent(AppEvent.CTX_ATTACHMENTS_CLEAR))

    def clear_silent(self):
        """Clear attachments list without confirmation"""
        # delete all from attachments for current mode
        mode = self.window.core.config.get('mode')
        self.window.core.attachments.delete_all(
            mode=mode,
            remove_local=False,
        )

    def open_add(self):
        """Open add attachment file dialog"""
        last_dir = self.window.core.config.get_last_used_dir()
        mode = self.window.core.config.get('mode')
        dialog = QFileDialog(self.window)
        dialog.setDirectory(last_dir)
        dialog.setFileMode(QFileDialog.ExistingFiles)
        if dialog.exec():
            files = dialog.selectedFiles()
            if files:
                self.window.core.config.set_last_used_dir(
                    os.path.dirname(files[0]),
                )
                for path in files:
                    # build attachment object
                    basename = os.path.basename(path)
                    attachment = self.window.core.attachments.new(
                        mode=mode,
                        name=basename,
                        path=path,
                        auto_save=False,
                    )

            # save attachments and update attachments list
            self.window.core.attachments.save()
            self.update()

    def open_add_url(self):
        """Open add attachment URL dialog"""
        self.window.ui.dialog['url'].id = "attachment"
        self.window.ui.dialog['url'].current = ""
        self.window.ui.dialog['url'].init()
        self.window.ui.dialog['url'].resize(800, 400)
        self.window.ui.dialog['url'].show()

    def attach_url(self):
        """Attach attachment URL"""
        result, loader, input_params, input_config = self.window.core.idx.ui.loaders.handle_options(
            self.window.ui.nodes["dialog.url.loader"],
            "dialog.url.loader.option",
            "dialog.url.loader.config",
        )
        extra = {
            "loader": loader,
            "input_params": input_params,
            "input_config": input_config,
        }
        provider = self.window.core.idx.indexing.get_loader(loader)
        if provider:
            mode = self.window.core.config.get('mode')
            name = provider.get_external_id(input_params)
            attachment = self.window.core.attachments.new(
                mode=mode,
                name=name,
                path=name,
                auto_save=False,
                type=AttachmentItem.TYPE_URL,
                extra=extra,
            )
            self.window.core.attachments.save()
            self.update()
            self.window.ui.dialog['url'].close()

    def add_url(self, url: str):
        """
        Add URL

        :param url: URL
        """
        if not url:
            return
        mode = self.window.core.config.get('mode')
        try:
            domain = urlparse(url).netloc
        except Exception as e:
            domain = os.path.basename(url)
        attachment = self.window.core.attachments.new(
            mode=mode,
            name=domain,
            path=url,
            auto_save=False,
            type=AttachmentItem.TYPE_URL,
        )
        self.window.core.attachments.save()
        self.update()
        self.window.ui.dialog['url'].close()

    def open_dir(self, mode: str, idx: int):
        """
        Open in directory

        :param mode: mode
        :param idx: index
        """
        path = self.get_path_by_idx(
            mode=mode,
            idx=idx,
        )
        if path is not None and path != '' and os.path.exists(path):
            self.window.controller.files.open_dir(
                path=path,
                select=True,
            )

    def open(self, mode: str, idx: int):
        """
        Open attachment

        :param mode: mode
        :param idx: index
        """
        path = self.get_path_by_idx(
            mode=mode,
            idx=idx,
        )
        if path is not None and path != '' and os.path.exists(path):
            self.window.controller.files.open(
                path=path,
            )

    def get_path_by_idx(self, mode: str, idx: int) -> str:
        """
        Get path by index

        :param mode: mode
        :param idx: index
        :return: path
        """
        file_id = self.window.core.attachments.get_id_by_idx(
            mode=mode,
            idx=idx,
        )
        data = self.window.core.attachments.get_by_id(
            mode=mode,
            id=file_id,
        )
        if data is None:
            return ''
        return data.path

    def get_by_idx(self, mode: str, idx: int) -> str:
        """
        Get attachment by index

        :param mode: mode
        :param idx: index
        :return: path
        """
        file_id = self.window.core.attachments.get_id_by_idx(
            mode=mode,
            idx=idx,
        )
        data = self.window.core.attachments.get_by_id(
            mode=mode,
            id=file_id,
        )
        return data

    def has(self, mode: str) -> bool:
        """
        Return True if current mode has attachments

        :param mode: mode to check
        :return: True if has attachments
        """
        return self.window.core.attachments.has(mode)

    def get_download_path(self, file_name: str) -> str:
        """
        Get file download path

        :param file_name: file name
        :return: download directory
        """
        if self.window.core.config.has("download.dir") and self.window.core.config.get("download.dir") != "":
            path = os.path.join(
                self.window.core.config.get_user_dir('data'),
                self.window.core.config.get("download.dir"),
                file_name,
            )
        else:
            path = os.path.join(
                self.window.core.config.get_user_dir('data'),
                file_name,
            )
        return str(path)

    def download(
            self,
            file_id: str,
            ext: Optional[str] = None
    ) -> Optional[str]:
        """
        Download file

        :param file_id: file id to download (id in OpenAI API)
        :param ext: file extension to add (optional)
        :return: path to downloaded file
        """
        try:
            # get file info from assistant API
            data = self.window.core.gpt.store.get_file(file_id)
            if data is None:
                return

            # prepare path to download file
            data.filename = os.path.basename(data.filename)

            # add extension if provided
            if ext is not None:
                data.filename = data.filename + ext

            # prepare path to downloaded file
            path = self.get_download_path(data.filename)

            # create download directory if not exists
            directory = os.path.dirname(path)
            if not os.path.exists(directory):
                os.makedirs(directory, exist_ok=True)

            # check if file exists, if yes, append timestamp prefix
            if os.path.exists(path):
                # append timestamp prefix to filename
                filename = f'{datetime.now().strftime("%Y%m%d%H%M%S")}_{data.filename}'
                path = self.get_download_path(filename)

            # download file
            self.window.core.gpt.store.download(
                file_id=file_id,
                path=path,
            )
            return path  # return path to downloaded file
        except Exception as e:
            self.window.core.debug.log(e)
            self.window.ui.dialogs.alert(e)

    def toggle_send_clear(self, value: bool):
        """
        Toggle send clear

        :param value: value of the checkbox
        """
        self.window.core.config.set('attachments_send_clear', value)

    def toggle_capture_clear(self, value: bool):
        """
        Toggle capture clear

        :param value: value of the checkbox
        """
        self.window.core.config.set('attachments_capture_clear', value)

    def toggle_auto_index(self, value: bool):
        """
        Toggle auto index

        :param value: value of the checkbox
        """
        self.window.core.config.set('attachments_auto_index', value)

    def is_capture_clear(self) -> bool:
        """
        Return True if capture clear is enabled

        :return: True if capture clear is enabled
        """
        if not self.window.core.config.has('attachments_capture_clear'):
            self.window.core.config.set('attachments_capture_clear', False)
        return self.window.core.config.get('attachments_capture_clear')

    def is_send_clear(self) -> bool:
        """
        Return True if send clear is enabled

        :return: True if send clear is enabled
        """
        if not self.window.core.config.has('attachments_send_clear'):
            self.window.core.config.set('attachments_send_clear', False)
        return self.window.core.config.get('attachments_send_clear')

    def from_clipboard_image(self, image):
        """
        Handle image from clipboard

        :param image: image data
        """
        now = datetime.now()
        dt = now.strftime("%Y-%m-%d_%H-%M-%S")
        name = 'clipboard-' + dt
        path = os.path.join(self.window.controller.painter.common.get_capture_dir(), name + '.png')
        image.save(path, "PNG")
        self.from_clipboard_url(path)

    def from_clipboard_url(self, url: str, all: bool = False):
        """
        Handle image from clipboard url

        :param url: file url
        :param all: all images
        """
        if not os.path.exists(url):
            return

        if not all:
            image_ext = ['.png', '.jpg', '.jpeg', '.gif', '.bmp', '.tiff']
            ext = os.path.splitext(url)[1].lower()
            if ext not in image_ext:
                return

        mode = self.window.core.config.get('mode')
        title = "Clipboard image"
        self.window.core.attachments.new(mode, title, url, False)
        self.window.core.attachments.save()
        self.window.controller.attachment.update()
        event = KernelEvent(KernelEvent.STATUS, {
            'status': trans("painter.capture.manual.captured.success") + ' ' + os.path.basename(url),
        })
        self.window.dispatch(event)

    def from_clipboard_text(self, text: str, all: bool = False):
        """
        Handle text from clipboard

        :param text: text from clipboard
        :param all: all text
        """
        if all:
            if os.path.exists(text):
                self.from_clipboard_url(text)
        else:
            if self.window.controller.chat.vision.allowed():
                # check if pasted text is local image path
                image_ext = ['.png', '.jpg', '.jpeg', '.gif', '.bmp', '.tiff']
                ext = os.path.splitext(text)[1].lower()
                if ext not in image_ext:
                    self.from_clipboard_url(text)
                    return
                if os.path.exists(text):
                    self.from_clipboard_url(text)

    def clipboard_has_attachment(self) -> bool:
        """
        Check if clipboard has attachment

        :return: True if clipboard has attachment
        """
        clipboard = QApplication.clipboard()
        source = clipboard.mimeData()
        if source.hasImage():
            image = source.imageData()
            if isinstance(image, QImage):
                return True
        elif source.hasUrls():
            urls = source.urls()
            for url in urls:
                if url.isLocalFile():
                    local_path = url.toLocalFile()
                    if os.path.exists(local_path):
                        return True
        elif source.hasText():
            text = source.text()
            if os.path.exists(text):
                return True
        return False

    def lock(self):
        """Lock attachment (disable clear)"""
        self.is_lock = True

    def unlock(self):
        """Unlock attachment (enable clear)"""
        self.is_lock = False

    def is_locked(self) -> bool:
        """
        Return True if attachment is locked

        :return: True if attachment is locked
        """
        return self.is_lock

    def clear_allowed(self, ctx: CtxItem) -> bool:
        """
        Check if clear is allowed

        :param ctx: context item
        :return: True if clear is allowed
        """
        if not ctx.cmds:
            return True
        for item in ctx.cmds:
            if "cmd" in item:
                if item["cmd"] in ["analyze_image_attachment"]:
                    return False
        return True

    def reload(self):
        """Reload attachments"""
        self.setup()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 18:00:00                  #
# ================================================== #

import base64
import os

from pygpt_net.core.types import (
    MODE_AUDIO,
)
from pygpt_net.core.bridge.context import MultimodalContext, BridgeContext
from pygpt_net.core.events import KernelEvent
from pygpt_net.item.ctx import CtxItem


class Audio:
    def __init__(self, window=None):
        """
        Chat audio controller

        :param window: Window instance
        """
        self.window = window
        self.audio_file = "chat_output.wav"
        self.tmp_input = False
        self.tmp_output = False

    def setup(self):
        """Set up UI"""
        pass

    def update(self):
        """Update input/output audio"""
        mode = self.window.core.config.get("mode")
        if mode == MODE_AUDIO:
            if not self.window.controller.audio.is_output_enabled():
                self.window.controller.audio.enable_output()
                self.tmp_output = True
            else:
                self.tmp_output = False
            if not self.window.controller.audio.is_input_enabled():
                self.window.controller.audio.enable_input()
                self.tmp_input = True
            else:
                self.tmp_input = False
        else:
            if self.tmp_output:
                self.window.controller.audio.disable_output()
            if self.tmp_input:
                self.window.controller.audio.disable_input()

    def handle_output(self, ctx: CtxItem):
        """
        Handle output audio

        :param ctx: Context item
        """
        wav_path = os.path.join(self.window.core.config.get_user_path(), self.audio_file)
        if ctx.is_audio and ctx.audio_output and ctx.audio_read_allowed():
            wav_bytes = base64.b64decode(ctx.audio_output)
            with open(wav_path, "wb") as f:
                f.write(wav_bytes)
            self.window.controller.audio.play_chat_audio(wav_path)

    def handle_input(self, path: str):
        """
        Handle input audio

        :param path: audio file path
        """
        multimodal_ctx = MultimodalContext()
        with open(path, "rb") as f:
            multimodal_ctx.audio_data = f.read()
        multimodal_ctx.is_audio_input = True

        bridge_ctx = BridgeContext()
        bridge_ctx.prompt = self.window.ui.nodes['input'].toPlainText()  # attach text input
        bridge_ctx.multimodal_ctx = multimodal_ctx
        event = KernelEvent(KernelEvent.INPUT_USER, {
            'context': bridge_ctx,
            'extra': {},
        })
        self.window.dispatch(event)

    def enabled(self) -> bool:
        """
        Check if audio mode is enabled

        :return: bool True if enabled
        """
        return self.window.core.config.get("mode") == MODE_AUDIO

